# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import json
import logging
import threading
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict

from rich.columns import Columns
from rich.console import Console
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    TextColumn,
)
from rich.table import Table
from rich.text import Text

from ...core.context import get_context
from ...core.orchestrator.observer import Observer
from ...core.orchestrator.termination import RunCancelError, SessionCancelError
from ...core.types import Action, ModelSettings, SessionOutcomeStatus
from ...utils.paths import RunPaths, get_run_paths, get_session_paths
from ...utils.settings import get_settings
from .recap import RunRecapMixin
from .session_ledger import SessionLedger


class _DurationColumn(ProgressColumn):
    def render(self, task) -> Text:
        duration = task.fields.get("duration")
        if isinstance(duration, (int, float)):
            elapsed = duration
        else:
            elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return Text("-:--:--", style="progress.elapsed")
        delta = timedelta(seconds=max(0, int(elapsed)))
        return Text(str(delta), style="progress.elapsed")


class _CountColumn(ProgressColumn):
    def render(self, task) -> Text:
        unit = task.fields.get("unit") or ""
        hide_total = bool(task.fields.get("hide_total"))
        total = task.total
        completed = int(task.completed or 0)
        if hide_total or unit == "steps":
            text = f"{completed} {unit}".strip()
        elif total is None:
            text = f"{completed} {unit}".strip()
        else:
            text = f"{completed}/{int(total)} {unit}".strip()
        return Text(text, style="progress.remaining")


class ConsoleLoggerObserver(Observer, RunRecapMixin):
    _MAX_REUSE_DURATION_SECONDS = 7 * 24 * 60 * 60
    _MAX_VISIBLE_SESSIONS = 10

    def __init__(self, console: Console | None = None) -> None:
        self._console = console or Console()
        self._lock = threading.Lock()
        self._ledger = SessionLedger()
        self._start_time: datetime | None = None
        self._progress: Progress | None = None
        self._run_task_id: int | None = None
        self._session_tasks: Dict[str, tuple[int, int | None]] = {}
        self._completed_session_tasks: list[int] = []
        self._run_config = None

    def on_run_start(self, run_config) -> None:
        if not self._enabled(logging.INFO):
            return
        self._run_config = run_config
        self._start_time = datetime.now()
        run_ctx = get_context()
        run_id = run_ctx.run_id

        # Display OTEL configuration if enabled
        settings = get_settings()
        if settings.otel_enabled:
            import os

            endpoint = os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT", "not set")
            protocol = os.getenv("OTEL_EXPORTER_OTLP_PROTOCOL", "http/protobuf")
            service_name = os.getenv("OTEL_SERVICE_NAME", "exgentic")
            record_content = "yes" if settings.otel_record_content else "no"

            otel_lines = [
                "[bold cyan]📊 OpenTelemetry Tracing ENABLED[/bold cyan]",
                f"[bold]Service Name:[/bold]     {service_name}",
                f"[bold]Collector:[/bold]        {endpoint}",
                f"[bold]Protocol:[/bold]         {protocol}",
                f"[bold]Record Content:[/bold]   {record_content}",
            ]
            otel_body = "\n".join(otel_lines)
            self._print(Panel(otel_body, border_style="cyan", padding=(1, 2), title="OpenTelemetry"))

        lines = [f"[bold]Run:[/bold] [cyan]{run_id}[/cyan]"]
        overrides = {}
        if run_config.max_steps != 100:
            overrides["max_steps"] = str(run_config.max_steps)
        if run_config.max_actions != 100:
            overrides["max_actions"] = str(run_config.max_actions)
        if run_config.max_workers is not None:
            overrides["max_workers"] = str(run_config.max_workers)
        if overrides:
            for key in sorted(overrides):
                lines.append(f"[bold]{key}:[/bold] {overrides[key]}")
        body = "\n".join(lines) + "\n"
        title = Text("EXGENTIC", style="bold magenta")
        self._print(Panel(body, border_style="magenta", padding=(1, 2), title=title))
        config_panels = self._build_config_panels()
        if config_panels:
            self._print(config_panels)
        self._start_progress(run_config)

    def on_session_start(self, session, agent, observation) -> None:
        if not self._enabled(logging.INFO):
            return
        session_id = session.session_id
        session_number = self._ledger.register(session_id)
        self._start_session_progress(session_id, session_number, agent, session)

    def on_react_success(self, session, action) -> None:
        if not isinstance(action, Action):
            return
        if not self._enabled(logging.INFO):
            return
        session_id = session.session_id
        self._ledger.increment_steps(session_id)
        self._advance_session_progress(session_id)

    def on_session_success(self, session, score, agent) -> None:
        if not self._enabled(logging.INFO):
            return
        session_id = session.session_id
        session_number = self._ledger.get_number(session_id)
        limit_reached = False
        try:
            limit_reached = bool(score.session_metadata.get("limit_reached"))
        except Exception:
            limit_reached = False
        status = None
        if limit_reached and not (score.is_finished is True and bool(score.success)):
            status = SessionOutcomeStatus.LIMIT_REACHED
        outcome = self._format_outcome(
            status=status,
            success=bool(score.success),
            is_finished=score.is_finished,
        )
        if outcome == "success":
            link = self._format_path_link(session.paths.root)
            desc = f"Session {session_number} ✔ success ({link})"
            color = "green"
        else:
            link = self._format_path_link(session.paths.root)
            desc = f"Session {session_number} ⏹ {outcome} ({link})"
            color = "yellow"
        self._update_session_progress_description(session_id, f"[{color}]{desc}[/{color}]")
        self._stop_session_progress(session_id)
        self._advance_run_progress()

    def on_session_scoring(self, session) -> None:
        if not self._enabled(logging.INFO):
            return
        session_id = session.session_id
        session_number = self._ledger.get_number(session_id)
        link = self._format_path_link(session.paths.root)
        desc = f"Session {session_number} ⏳ scoring ({link})"
        self._update_session_progress_description(session_id, f"[yellow]{desc}[/yellow]")
        if self._progress is None:
            self._print(Text.from_markup(f"[yellow]{desc}[/yellow]"))
        else:
            self._progress.refresh()

    def on_session_error(self, session, error) -> None:
        session_id = session.session_id if session else None
        session_root = None
        if session is not None:
            session_root = session.paths.root
        elif session_id is not None:
            try:
                ctx = get_context()
                session_root = RunPaths.from_context(ctx).session(session_id).root
            except RuntimeError:
                pass
        session_number = self._ledger.get_number(session_id)
        if not self._enabled(logging.INFO):
            return
        if isinstance(error, (SessionCancelError, RunCancelError, KeyboardInterrupt)):
            desc = f"[yellow]Session {session_number} ⏹ cancelled"
        else:
            desc = f"[red]Session {session_number} ✖ error"
        if session_root is not None:
            link = self._format_path_link(session_root)
            desc = f"{desc} ({link})"
        if desc.startswith("[red]"):
            desc = f"{desc}[/red]"
        else:
            desc = f"{desc}[/yellow]"
        self._update_session_progress_description(session_id, desc)
        self._stop_session_progress(session_id)
        self._advance_run_progress()

    def on_session_reuse(self, session_results) -> None:
        if not self._enabled(logging.INFO):
            return
        session_id = session_results.session_id
        session_number = self._ledger.mark_reuse(session_id)
        steps = session_results.steps
        status = session_results.status
        execution_time = session_results.execution_time
        if (
            not isinstance(execution_time, (int, float))
            or execution_time < 0
            or execution_time > self._MAX_REUSE_DURATION_SECONDS
        ):
            execution_time = None
        outcome = self._format_outcome(
            status=status,
            success=session_results.success,
            is_finished=session_results.is_finished,
        )
        detail_parts = ["↺ reused", outcome]
        detail_text = " ".join(detail_parts)
        session_root = get_session_paths(session_id).root
        link = self._format_path_link(session_root)
        desc = f"[yellow]Session {session_number} {detail_text} " f"({link})[/yellow]"
        total = steps if isinstance(steps, int) and steps > 0 else 1
        self._add_completed_session_task(
            desc,
            total=total,
            duration=execution_time if isinstance(execution_time, (int, float)) else None,
        )
        self._advance_run_progress()

    def on_run_success(self, results, run_config) -> None:
        self._stop_progress()
        self._print_recap()

    def on_run_error(self, error) -> None:
        self._stop_progress()
        self._print_recap()

    def _print_recap(self) -> None:
        if not self._enabled(logging.INFO):
            return
        data = self._load_recap_data(get_run_paths().results, self._start_time)
        if data is None:
            return
        table = Table(show_header=False, box=None, pad_edge=False)
        table.add_row(
            "[bold]Sessions[/bold]",
            f"{data.total_sessions} (successes: {data.successful_sessions})",
        )
        if data.success_rate is not None:
            table.add_row(
                "[bold]Success %[/bold]",
                f"{data.success_rate:.2%}",
            )
        if data.finished_sessions is not None:
            table.add_row(
                "[bold]Finished[/bold]",
                f"{data.finished_sessions}",
            )
        table.add_row("[bold]Avg steps[/bold]", f"{data.average_steps}")
        self._print(Panel(table, border_style="magenta", title="Recap"))

        cost_table = Table(show_header=False, box=None, pad_edge=False)
        cost_table.add_row(
            "[bold]Run[/bold]",
            f"{self._format_money(data.run_cost)}",
        )
        cost_table.add_row(
            "[bold]Avg agent[/bold]",
            f"{self._format_money(data.avg_agent_cost)}",
        )
        self._print(Panel(cost_table, border_style="magenta", title="Costs"))

        results_table = Table(show_header=False, box=None, pad_edge=False)
        results_table.add_row("[bold]Results[/bold]", f"{data.results_path}")
        self._print(Panel(results_table, border_style="magenta", title="Results"))

    def _enabled(self, level: int) -> bool:
        configured = logging._nameToLevel.get(get_settings().log_level.upper(), logging.INFO)
        return level >= configured

    def _format_value(self, value) -> str:
        if isinstance(value, (dict, list, tuple)):
            try:
                return json.dumps(value, ensure_ascii=True)
            except TypeError:
                return str(value)
        return str(value)

    def _build_config_panels(self):
        if self._run_config is None:
            return None
        bench_overrides = dict(self._run_config.benchmark_kwargs or {})
        agent_overrides = dict(self._run_config.agent_kwargs or {})
        model_value = self._run_config.model or agent_overrides.get("model") or "unknown"
        model_settings = agent_overrides.pop("model_settings", None)
        agent_overrides = {
            "model": str(model_value),
            **agent_overrides,
        }
        if model_settings:
            if hasattr(model_settings, "model_dump"):
                model_settings = model_settings.model_dump(exclude_none=True)
            if isinstance(model_settings, dict):
                default_settings = ModelSettings().model_dump(exclude_none=True)
                for key, value in model_settings.items():
                    if value is None:
                        continue
                    if default_settings.get(key) == value:
                        continue
                    agent_overrides[f"model.{key}"] = value

        bench_name = self._run_config.benchmark
        agent_name = self._run_config.agent
        bench_panel = self._build_config_panel(f"Benchmark: {bench_name}", bench_overrides, border_style="cyan")
        agent_panel = self._build_config_panel(f"Agent: {agent_name}", agent_overrides, border_style="green")
        console_width = self._console.width
        gap = 2
        panel_width = max(20, (console_width - gap) // 2)
        bench_panel.width = panel_width
        agent_panel.width = panel_width
        return Columns([bench_panel, agent_panel], equal=True, expand=True)

    def _build_config_panel(
        self,
        title: str,
        overrides: Dict[str, str],
        *,
        border_style: str = "magenta",
    ) -> Panel:
        table = Table(show_header=False, box=None, pad_edge=False)
        if overrides:
            for key in sorted(overrides):
                table.add_row(f"[bold]{key}[/bold]", self._format_value(overrides[key]))
        else:
            table.add_row("[dim]no overrides[/dim]", "")
        return Panel(table, border_style=border_style, title=title)

    def _format_money(self, value: float | None) -> str:
        if value is None:
            return "-"
        return f"${value:.1f}"

    def _format_score(self, value) -> str:
        if value is None:
            return "-"
        if isinstance(value, (int, float)):
            return f"{value:.2f}"
        return str(value)

    def _format_outcome(
        self,
        *,
        status=None,
        success: bool | None = None,
        is_finished: bool | None = None,
    ) -> str:
        if status is not None:
            return str(status)
        if is_finished is False:
            return "unfinished"
        if is_finished is True:
            return "success" if success else "unsuccessful"
        return "unknown"

    @staticmethod
    def _format_path_link(path, *, max_len: int = 80) -> str:
        text = str(path)
        if len(text) > max_len and max_len > 3:
            text = "..." + text[-(max_len - 3) :]
        try:
            target = str(Path(path).resolve())
        except Exception:
            return text
        return f"[link={target}]{text}[/link]"

    def _print(self, renderable) -> None:
        with self._lock:
            if self._progress is not None:
                self._progress.console.print(renderable)
            else:
                self._console.print(renderable)

    def _start_progress(self, run_config) -> None:
        total = None
        if run_config is not None:
            if run_config.task_ids:
                total = len(run_config.task_ids)
                if run_config.num_tasks is not None:
                    total = min(total, int(run_config.num_tasks))
            elif run_config.num_tasks is not None:
                total = int(run_config.num_tasks)
        self._progress = Progress(
            SpinnerColumn(),
            TextColumn("[bold]{task.description}[/bold]"),
            BarColumn(bar_width=None),
            _CountColumn(),
            _DurationColumn(),
            console=self._console,
            transient=False,
        )
        self._progress.start()
        self._run_task_id = self._progress.add_task("Run", total=total if total else None, unit="sessions")

    def _start_session_progress(self, session_id: str, session_number: int, agent, session) -> None:
        if self._progress is None:
            return
        total = agent.max_steps
        task_id = self._progress.add_task(
            f"Session {session_number} ({self._format_path_link(session.paths.root)})",
            total=total if total else None,
            unit="steps",
            hide_total=True,
        )
        self._session_tasks[session_id] = (task_id, total)

    def _update_session_progress_description(self, session_id: str, description: str) -> None:
        if self._progress is None:
            return
        entry = self._session_tasks.get(session_id)
        if entry is None:
            return
        task_id, _ = entry
        self._progress.update(task_id, description=description)

    def _add_completed_session_task(
        self,
        description: str,
        total: int,
        duration: float | None = None,
    ) -> None:
        if self._progress is None:
            return
        task_id = self._progress.add_task(
            description,
            total=total,
            unit="steps",
            duration=duration,
        )
        self._progress.update(task_id, completed=total)
        self._progress.stop_task(task_id)
        self._track_completed_session_task(task_id)

    def _advance_session_progress(self, session_id: str) -> None:
        if self._progress is None:
            return
        entry = self._session_tasks.get(session_id)
        if entry is None:
            return
        task_id, _ = entry
        self._progress.update(task_id, advance=1)

    def _stop_session_progress(self, session_id: str) -> None:
        if self._progress is None:
            return
        entry = self._session_tasks.pop(session_id, None)
        if entry is None:
            return
        task_id, total = entry
        steps = self._ledger.get_steps(session_id)
        if steps > 0:
            self._progress.update(task_id, total=steps, completed=steps)
        else:
            self._progress.update(task_id, completed=0)
        self._progress.stop_task(task_id)
        self._track_completed_session_task(task_id)

    def _track_completed_session_task(self, task_id: int) -> None:
        if self._progress is None:
            return
        self._completed_session_tasks.append(task_id)
        excess = len(self._completed_session_tasks) - self._MAX_VISIBLE_SESSIONS
        if excess <= 0:
            return
        for _ in range(excess):
            old_id = self._completed_session_tasks.pop(0)
            if old_id == self._run_task_id:
                continue
            try:
                self._progress.remove_task(old_id)
            except Exception:
                continue

    def _advance_run_progress(self) -> None:
        if self._progress is None or self._run_task_id is None:
            return
        self._progress.update(self._run_task_id, advance=1)

    def _stop_progress(self) -> None:
        if self._progress is None:
            return
        self._progress.stop()
        self._progress = None
        self._run_task_id = None
        self._session_tasks = {}



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""ReplayAgent — replays a recorded trajectory against a benchmark.

Usage::

    evaluate(benchmark="gsm8k", agent="replay", agent_kwargs={"recording": "path/to/recording"})

A *recording* is a directory containing:
    trajectory.jsonl   — the recorded action/observation events
    session.json       — session manifest (task, context, actions schema)

These files are produced automatically by ``exgentic evaluate`` (under
``outputs/<run_id>/sessions/<session_id>/``).
"""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any, ClassVar

from ...core.agent import Agent
from ...core.agent_instance import AgentInstance
from ...core.types import Action, Observation


class ReplayAgentInstance(AgentInstance):
    """Replays recorded actions from a trajectory file."""

    def __init__(
        self,
        *,
        session_id: str,
        trajectory_path: Path,
    ) -> None:
        super().__init__(session_id=session_id)
        self._actions = self._load_actions(trajectory_path)
        self._step = 0

    def start(self, task, context, actions):
        """Receive work payload and build the action_types lookup."""
        super().start(task, context, actions)
        self._action_types = {at.name: at for at in self.actions}

    @staticmethod
    def _load_actions(trajectory_path: Path) -> list[dict]:
        """Extract action events from a trajectory JSONL file."""
        actions = []
        with open(trajectory_path, encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                event = json.loads(line)
                if event.get("event") == "action":
                    actions.append(event["action"])
        return actions

    def react(self, observation: Observation | None) -> Action | None:
        if self._step >= len(self._actions):
            return None  # No more recorded actions — signal done

        action_data = self._actions[self._step]
        self._step += 1

        name = action_data.get("name", "")
        arguments = action_data.get("arguments", {})

        action_type = self._action_types.get(name)
        if action_type is None:
            from ...core.actions import build_unknown_action

            return build_unknown_action(name, arguments)

        return action_type.build_action(arguments)

    def close(self) -> None:
        pass


class ReplayAgent(Agent):
    """Agent that replays pre-recorded actions from a trajectory file."""

    display_name: ClassVar[str] = "Replay Agent"
    slug_name: ClassVar[str] = "replay"
    recording: str  # Path to the recording directory (or trajectory.jsonl file)
    runner: str | None = "direct"  # No external deps — run in host process

    @classmethod
    def get_instance_class(cls):
        return ReplayAgentInstance

    def get_instance_kwargs(
        self,
        session_id: str,
    ) -> dict[str, Any]:
        recording_path = Path(self.recording)
        if recording_path.is_file():
            trajectory_path = recording_path
        else:
            trajectory_path = recording_path / "trajectory.jsonl"
        return {
            "session_id": session_id,
            "trajectory_path": trajectory_path,
        }



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import json
import secrets
from abc import ABC, abstractmethod
from logging import Logger
from typing import Any, Dict, List, Optional

from ..observers.logging import get_logger
from ..utils.cost import CostReport
from ..utils.paths import SessionPaths
from .types import Action, ActionType, Observation, SessionScore


class Session(ABC):
    """Session interface - represents one task execution in one environment."""

    def __init__(self) -> None:
        # Persist configuration once per session if provided by subclass.
        self.save_config()
        self.save_manifest()

    @property
    def session_id(self) -> str:
        """Returns a unique id for the session."""
        if not hasattr(self, "_session_id"):
            self._session_id = secrets.token_hex(4)
        return self._session_id

    @property
    def paths(self) -> SessionPaths:
        """Convenience wrapper for all filesystem paths for this session."""
        if not hasattr(self, "_paths"):
            from .context import try_get_context

            ctx = try_get_context()
            if ctx is not None:
                self._paths = SessionPaths(
                    session_id=self.session_id,
                    run_id=ctx.run_id,
                    output_dir=ctx.output_dir,
                )
            else:
                self._paths = SessionPaths(session_id=self.session_id, run_id="default", output_dir="outputs")
        return self._paths

    @property
    def logger(self) -> Logger:
        if not hasattr(self, "_logger"):
            self._logger = get_logger(f"Session_{self.session_id}", str(self.paths.session_log))
        return self._logger

    def get_config(self) -> Dict[str, Any]:
        """Return a serializable configuration dict for this session."""
        return {}

    def save_config(self) -> None:
        """Persist the session configuration to the standard config path."""
        config = self.get_config()
        config_path = self.paths.benchmark_config
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, "w", encoding="utf-8") as f:
            json.dump(config, f, ensure_ascii=False, indent=2)

    def save_results(self, payload: Dict[str, Any]) -> None:
        """Persist a results payload to the standard per-session results path."""
        results_path = self.paths.benchmark_results
        results_path.parent.mkdir(parents=True, exist_ok=True)
        with open(results_path, "w", encoding="utf-8") as f:
            json.dump(payload, f, ensure_ascii=False, indent=2)

    def save_standard_results(self, score: SessionScore) -> None:
        """Persist the minimal standardized results payload (score/success)."""
        self.save_results({"score": score.score, "success": bool(score.success)})

    def save_manifest(self) -> None:
        """Persist task/context/actions metadata for the session."""
        actions_payload = []
        for action_type in self.actions:
            action_entry = {
                "name": action_type.name,
                "description": action_type.description,
                "is_finish": bool(action_type.is_finish),
                "is_message": bool(action_type.is_message),
                "is_hidden": bool(action_type.is_hidden),
            }
            schema = action_type.arguments.model_json_schema()  # type: ignore[attr-defined]
            action_entry["arguments_schema"] = schema
            actions_payload.append(action_entry)

        manifest = {
            "run_id": self.paths.run_id,
            "session_id": self.session_id,
            "task": self.task,
            "context": self.context,
            "actions": actions_payload,
            "task_id": self.task_id,
        }
        path = self.paths.session_manifest
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w", encoding="utf-8") as f:
            json.dump(manifest, f, ensure_ascii=False, indent=2)

    def get_cost(self) -> CostReport:
        """Estimated session cost; default 0.0."""
        return CostReport.initialize_empty()

    @property
    def task_id(self) -> str:
        """Task identifier."""
        return ""

    @property
    @abstractmethod
    def task(self) -> str:
        """Task description - benchmark defines what work to do."""
        pass

    @property
    @abstractmethod
    def context(self) -> Dict[str, Any]:
        """Task context - benchmark provides necessary information."""
        pass

    @property
    @abstractmethod
    def actions(self) -> List[ActionType]:
        """Available actions - benchmark defines action space."""
        pass

    @abstractmethod
    def start(self) -> Optional[Observation]:
        """Current observation - session maintains state."""
        pass

    @abstractmethod
    def step(self, action: Action) -> Optional[Observation]:
        """Execute action - session controls execution, returns None when done."""
        pass

    @abstractmethod
    def done(self) -> bool:
        """Check completion - session knows when task is finished."""
        pass

    @abstractmethod
    def score(self) -> Dict[str, Any]:
        """Evaluate performance - session/benchmark controls scoring.

        Should return at least two fields:

        "success" - True iff the session execution completed without an error
        "score" - The score of the session execution, should be None if an error occurred.
        """
        pass

    @abstractmethod
    def close(self):
        """Cleanup resources - session manages its own resources."""
        pass



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
import math
from typing import Any


def _format_value(value: Any) -> str:
    if value is None:
        return "-"
    if isinstance(value, float):
        if math.isnan(value):
            return "NaN"
        abs_val = abs(value)
        if abs_val >= 100 or value.is_integer():
            return f"{value:.2f}".rstrip("0").rstrip(".")
        if abs_val >= 1:
            return f"{value:.3f}".rstrip("0").rstrip(".")
        return f"{value:.6f}".rstrip("0").rstrip(".")
    return str(value)


def _format_payload(value: Any, limit: int = 4000) -> str:
    try:
        payload = json.dumps(value, ensure_ascii=False, indent=2, default=str)
    except Exception:
        payload = str(value)
    if len(payload) > limit:
        return payload[:limit] + "..."
    return payload


def _format_error_message(value: Any, limit: int = 4000) -> str:
    if isinstance(value, str):
        if len(value) > limit:
            return value[:limit] + "..."
        return value
    return _format_payload(value, limit=limit)


def _normalize_action_payload(payload: Any) -> dict:
    if payload is None:
        return {"name": None, "arguments": None}
    if isinstance(payload, dict):
        action_type = payload.get("type")
        if action_type in {"parallel", "sequential"}:
            actions = payload.get("actions") or []
            cleaned_actions = []
            if isinstance(actions, list):
                for item in actions:
                    if not isinstance(item, dict):
                        cleaned_actions.append({"name": str(item), "arguments": None})
                        continue
                    cleaned_actions.append(
                        {
                            "name": item.get("name"),
                            "arguments": item.get("arguments"),
                        }
                    )
            return {"mode": action_type, "actions": cleaned_actions}
        return {
            "name": payload.get("name"),
            "arguments": payload.get("arguments"),
        }
    return {"raw": payload}


def _normalize_observation_item(item: Any) -> Any:
    if item is None:
        return None
    if isinstance(item, dict):
        if "result" in item:
            result = item.get("result")
            if isinstance(result, dict):
                if "sender" in result and "message" in result:
                    return {
                        "sender": result.get("sender"),
                        "message": result.get("message"),
                    }
            return result
        if "sender" in item and "message" in item:
            return {"sender": item.get("sender"), "message": item.get("message")}
        return item
    return item


def _normalize_observation_payload(payload: Any) -> Any:
    if payload is None:
        return None
    if isinstance(payload, dict):
        if isinstance(payload.get("observations"), list):
            return [_normalize_observation_item(item) for item in payload["observations"]]
        return _normalize_observation_item(payload)
    if isinstance(payload, list):
        return [_normalize_observation_item(item) for item in payload]
    return payload



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

from click.testing import CliRunner
from exgentic import __version__
from exgentic.interfaces.cli.main import cli


def test_cli_version_long_flag():
    """Test that --version flag displays version and exits."""
    runner = CliRunner()
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert f"exgentic {__version__}" in result.output


def test_cli_version_short_flag():
    """Test that -V flag displays version and exits."""
    runner = CliRunner()
    result = runner.invoke(cli, ["-V"])
    assert result.exit_code == 0
    assert f"exgentic {__version__}" in result.output



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
from pathlib import Path

from exgentic.core.context import run_scope, try_get_context
from exgentic.core.types import RunConfig
from exgentic.integrations.litellm.cache import build_litellm_cache
from exgentic.integrations.litellm.config import configure_litellm
from exgentic.integrations.litellm.trace_logger import (
    FILE_ENV,
    AsyncTraceLogger,
    SyncTraceLogger,
    TraceLogger,
)
from exgentic.utils.settings import ExgenticSettings, resolve_cache_path


def test_resolve_cache_path_uses_base_dir_for_relative_paths() -> None:
    assert resolve_cache_path(".exgentic", ".litellm_cache") == ".exgentic/.litellm_cache"


def test_resolve_cache_path_keeps_absolute_paths() -> None:
    assert resolve_cache_path(".exgentic", "/tmp/litellm") == "/tmp/litellm"


def test_build_litellm_cache_resolves_relative_path_under_cache_dir(tmp_path) -> None:
    base_dir = tmp_path / "cache-root"
    settings = ExgenticSettings(
        cache_dir=str(base_dir),
        litellm_cache_dir=".litellm_cache",
    )
    cache = build_litellm_cache(settings)
    assert cache.cache.disk_cache.directory == str(base_dir / ".litellm_cache")


def test_run_config_to_session_config_preserves_cache_dir() -> None:
    run_config = RunConfig(
        benchmark="tau2",
        agent="tool_calling",
        cache_dir="/tmp/exgentic-cache",
    )
    session_config = run_config.to_session_config("task-1")
    assert session_config.cache_dir == "/tmp/exgentic-cache"


def test_run_scope_sets_and_restores_cache_env() -> None:
    before = try_get_context()
    with run_scope(
        run_id="cache-test",
        output_dir="./outputs",
        cache_dir="./cache",
    ):
        ctx = try_get_context()
        assert ctx is not None
        assert ctx.cache_dir == str(Path("./cache").resolve())
    after = try_get_context()
    if before is None:
        assert after is None
    else:
        assert after == before


def test_configure_litellm_always_registers_trace_logger_callbacks() -> None:
    import litellm

    original_callbacks = litellm.callbacks
    original_success = litellm.success_callback
    original_failure = litellm.failure_callback
    original_async_success = litellm._async_success_callback
    original_async_failure = litellm._async_failure_callback
    try:
        litellm.callbacks = []
        litellm.success_callback = []
        litellm.failure_callback = []
        litellm._async_success_callback = []
        litellm._async_failure_callback = []

        settings = ExgenticSettings(litellm_caching=False)
        configure_litellm(config=settings.to_litellm_config(), cache_only=False)
        configure_litellm(config=settings.to_litellm_config(), cache_only=False)

        assert any(isinstance(cb, SyncTraceLogger) for cb in litellm.success_callback)
        assert any(isinstance(cb, AsyncTraceLogger) for cb in litellm.success_callback)
        assert any(isinstance(cb, SyncTraceLogger) for cb in litellm.failure_callback)
        assert any(isinstance(cb, AsyncTraceLogger) for cb in litellm.failure_callback)
        assert sum(isinstance(cb, SyncTraceLogger) for cb in litellm.success_callback) == 1
        assert sum(isinstance(cb, AsyncTraceLogger) for cb in litellm.success_callback) == 1
        assert sum(isinstance(cb, SyncTraceLogger) for cb in litellm.failure_callback) == 1
        assert sum(isinstance(cb, AsyncTraceLogger) for cb in litellm.failure_callback) == 1
        assert any(isinstance(cb, SyncTraceLogger) for cb in litellm._async_success_callback)
        assert any(isinstance(cb, AsyncTraceLogger) for cb in litellm._async_success_callback)
        assert any(isinstance(cb, SyncTraceLogger) for cb in litellm._async_failure_callback)
        assert any(isinstance(cb, AsyncTraceLogger) for cb in litellm._async_failure_callback)
    finally:
        litellm.callbacks = original_callbacks
        litellm.success_callback = original_success
        litellm.failure_callback = original_failure
        litellm._async_success_callback = original_async_success
        litellm._async_failure_callback = original_async_failure


def test_trace_logger_callback_registered_and_writes_by_default(tmp_path, monkeypatch) -> None:
    import litellm

    original_callbacks = litellm.callbacks
    original_success = litellm.success_callback
    original_failure = litellm.failure_callback
    original_async_success = litellm._async_success_callback
    original_async_failure = litellm._async_failure_callback
    try:
        litellm.callbacks = []
        litellm.success_callback = []
        litellm.failure_callback = []
        litellm._async_success_callback = []
        litellm._async_failure_callback = []

        log_path = tmp_path / "trace.jsonl"
        monkeypatch.setenv(FILE_ENV, str(log_path))

        settings = ExgenticSettings(litellm_caching=False)
        configure_litellm(config=settings.to_litellm_config(), cache_only=False)

        registered = [cb for cb in litellm.success_callback if isinstance(cb, TraceLogger)]
        assert len(registered) == 2

        kwargs = {
            "model": "openai/gpt-4o-mini",
            "messages": [{"role": "user", "content": "Hi"}],
            "response_cost": 0.0,
        }
        response = {
            "usage": {"prompt_tokens": 1, "completion_tokens": 1, "total_tokens": 2},
            "choices": [
                {
                    "message": {"role": "assistant", "content": "ok"},
                    "finish_reason": "stop",
                }
            ],
        }
        registered[0].log_success_event(kwargs, response, None, None)

        assert log_path.exists()
        record = json.loads(log_path.read_text(encoding="utf-8").splitlines()[0])
        assert record["status"] == "success"
        assert record["model"] == "openai/gpt-4o-mini"
    finally:
        litellm.callbacks = original_callbacks
        litellm.success_callback = original_success
        litellm.failure_callback = original_failure
        litellm._async_success_callback = original_async_success
        litellm._async_failure_callback = original_async_failure



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
import os
import socket
import subprocess
import sys
import time
from pathlib import Path

import pytest
import requests
from exgentic.integrations.litellm import LitellmProxy


def _free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return int(s.getsockname()[1])


def _wait_for_proxy(url: str, timeout: float = 10.0) -> None:
    deadline = time.time() + timeout
    while time.time() < deadline:
        try:
            if requests.get(url, timeout=1.0).status_code == 200:
                return
        except Exception:
            pass
        time.sleep(0.1)
    raise RuntimeError("LiteLLM proxy did not become ready")


@pytest.mark.skipif(
    not (Path(sys.executable).parent / "litellm").exists(),
    reason="litellm CLI not installed in active venv",
)
def test_exgentic_proxy_executes_default_trace_callback_and_writes_trace(tmp_path, fake_openai_server) -> None:
    backend_port = fake_openai_server.server_address[1]
    backend_base = f"http://127.0.0.1:{backend_port}/v1"
    trace_path = tmp_path / "trace.jsonl"

    env = os.environ.copy()
    env.update(
        {
            "OPENAI_API_BASE": backend_base,
            "OPENAI_API_KEY": "test-key",  # pragma: allowlist secret
            "EXGENTIC_LLM_LOG_FILE": str(trace_path),
        }
    )

    with LitellmProxy(
        model="openai/gpt-4o-mini",
        env=env,
        startup_timeout=10.0,
    ) as proxy:
        response = requests.post(
            f"{proxy.base_url}/v1/chat/completions",
            json={
                "model": "openai/gpt-4o-mini",
                "messages": [{"role": "user", "content": "Hello"}],
            },
            timeout=10,
        )
        response.raise_for_status()

    assert trace_path.exists()
    rows = [json.loads(line) for line in trace_path.read_text().splitlines() if line]
    assert rows and rows[0]["status"] == "success"


@pytest.mark.skipif(
    not (Path(sys.executable).parent / "litellm").exists(),
    reason="litellm CLI not installed in active venv",
)
def test_raw_litellm_proxy_executes_litellm_settings_callbacks_and_writes_trace(tmp_path, fake_openai_server) -> None:
    backend_port = fake_openai_server.server_address[1]
    backend_base = f"http://127.0.0.1:{backend_port}/v1"
    proxy_port = _free_port()
    trace_path = tmp_path / "trace.jsonl"
    config_path = tmp_path / "litellm_config.json"

    callback_path = "exgentic.integrations.litellm.trace_logger.trace_logger"
    async_callback_path = "exgentic.integrations.litellm.trace_logger.async_trace_logger"
    config_path.write_text(
        json.dumps(
            {
                "model_list": [
                    {
                        "model_name": "openai/gpt-4o-mini",
                        "litellm_params": {"model": "openai/gpt-4o-mini"},
                    }
                ],
                "litellm_settings": {
                    "success_callback": [callback_path, async_callback_path],
                    "failure_callback": [callback_path, async_callback_path],
                },
            }
        ),
        encoding="utf-8",
    )

    env = os.environ.copy()
    env.update(
        {
            "OPENAI_API_BASE": backend_base,
            "OPENAI_API_KEY": "test-key",  # pragma: allowlist secret
            "EXGENTIC_LLM_LOG_FILE": str(trace_path),
        }
    )
    repo_root = Path(__file__).resolve().parents[4]
    src_path = repo_root / "src"
    env["PYTHONPATH"] = os.pathsep.join(
        [
            str(src_path),
            str(repo_root),
            env.get("PYTHONPATH", ""),
        ]
    )

    proc = subprocess.Popen(
        [
            str(Path(sys.executable).parent / "litellm"),
            "--config",
            str(config_path),
            "--port",
            str(proxy_port),
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        env=env,
    )
    try:
        _wait_for_proxy(f"http://127.0.0.1:{proxy_port}/v1/models")
        response = requests.post(
            f"http://127.0.0.1:{proxy_port}/v1/chat/completions",
            json={
                "model": "openai/gpt-4o-mini",
                "messages": [{"role": "user", "content": "Hello"}],
            },
            timeout=10,
        )
        response.raise_for_status()
    finally:
        proc.terminate()
        try:
            proc.wait(timeout=10)
        except subprocess.TimeoutExpired:
            proc.kill()

    assert trace_path.exists()
    rows = [json.loads(line) for line in trace_path.read_text().splitlines() if line]
    assert rows and rows[0]["status"] == "success"



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Thread-specific tests that verify actual thread isolation."""

from __future__ import annotations

import threading

from exgentic.adapters.runners import with_runner

from .conftest import Calculator


def test_runs_in_different_thread():
    calc = with_runner(Calculator, runner="thread", value=0)
    try:
        remote_tid = calc.thread_id()
        local_tid = threading.get_ident()
        assert remote_tid != local_tid
    finally:
        calc.close()


def test_runs_in_same_pid():
    import os

    calc = with_runner(Calculator, runner="thread", value=0)
    try:
        assert calc.pid() == os.getpid()
    finally:
        calc.close()


def test_close_joins_thread():
    calc = with_runner(Calculator, runner="thread", value=0)
    transport = object.__getattribute__(calc, "_transport")
    thread = transport._thread
    assert thread is not None and thread.is_alive()
    calc.close()
    assert not thread.is_alive()



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
from pathlib import Path
from typing import Optional

from ..base import BaseCLIConfig, BaseCLIWrapper, ExecutionBackend


class GeminiCLIConfig(BaseCLIConfig):
    backend_model: str
    gemini_model: str = "gemini-2.5-pro"
    env_key: str = "GEMINI_API_KEY"
    server_name: str = "environment"
    output_format: str = "text"
    approval_mode: str = "yolo"
    allowed_mcp_server_names: Optional[list[str]] = None
    allowed_tools: Optional[list[str]] = None
    image: str = "exgentic-gemini:dev"


class GeminiCLI(BaseCLIWrapper):
    """Lightweight wrapper for running the Gemini CLI headlessly."""

    def __init__(
        self,
        env: Optional[dict[str, str]] = None,
        log_path: Optional[Path] = None,
        config_dir: Optional[Path] = None,
        logger=None,
        runner: ExecutionBackend = ExecutionBackend.PROCESS,
    ) -> None:
        super().__init__(
            env=env,
            log_path=log_path,
            config_dir=config_dir,
            logger=logger,
            runner=runner,
        )
        self._settings_path: Optional[Path] = None
        self.config_prefix = "gemini_cli_"
        self.spawn_error_message = "Failed to start Gemini CLI"

    # Required hooks --------------------------------------------------

    def build_env(self, *, cfg_root: Path, prompt: str, config: GeminiCLIConfig) -> dict[str, str]:
        env = self.env.copy()
        env["GOOGLE_GEMINI_BASE_URL"] = config.provider_url

        api_key = env.get(config.env_key) or env.get("GEMINI_API_KEY") or "dummy-api-key"
        env[config.env_key] = api_key
        env["GEMINI_API_KEY"] = api_key

        env["HOME"] = str(cfg_root)
        return env

    def build_command(self, *, cfg_root: Path, prompt: str, config: GeminiCLIConfig) -> list[str]:
        gemini_cfg_dir = cfg_root / ".gemini"
        settings_path = gemini_cfg_dir / "settings.json"
        self._settings_path = settings_path

        # Rewrite localhost addresses to host gateway for container runners
        mcp_host = config.mcp_host
        from ..command_runner import ContainerRunner

        if isinstance(self.runner, ContainerRunner) and mcp_host in ("0.0.0.0", "127.0.0.1", "localhost"):
            mcp_host = self.runner.host_gateway

        gemini_cfg_dir.mkdir(parents=True, exist_ok=True)
        mcp_url = f"http://{mcp_host}:{config.mcp_port}/mcp"
        self._ensure_settings(settings_path, config.server_name, mcp_url)

        cmd: list[str] = [
            "gemini",
            "--model",
            config.gemini_model,
            "--output-format",
            config.output_format,
            "--approval-mode",
            config.approval_mode,
        ]
        if config.allowed_mcp_server_names:
            cmd.extend(["--allowed-mcp-server-names", *config.allowed_mcp_server_names])
        if config.allowed_tools:
            cmd.extend(["--allowed-tools", *config.allowed_tools])
        cmd.append(prompt)
        return cmd

    # Internal helpers -------------------------------------------------

    def _ensure_settings(self, settings_path: Path, server_name: str, mcp_url: str) -> None:
        settings: dict[str, object] = {}
        if settings_path.exists():
            try:
                with open(settings_path, encoding="utf-8-sig") as fh:
                    settings = json.load(fh)
            except Exception:
                settings = {}

        servers = settings.setdefault("mcpServers", {})
        if not isinstance(servers, dict):
            servers = {}
            settings["mcpServers"] = servers

        servers[server_name] = {"httpUrl": mcp_url, "trust": True}

        settings_path.parent.mkdir(parents=True, exist_ok=True)
        with open(settings_path, "w", encoding="utf-8") as fh:
            json.dump(settings, fh, indent=2)



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
import os
import shlex
import shutil
import subprocess
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional

from pydantic import BaseModel


class ExecutionBackend(str, Enum):
    """Execution backend for CLI runs."""

    PROCESS = "process"
    PODMAN = "podman"
    DOCKER = "docker"
    AUTO = "auto"


def resolve_container_backend() -> ExecutionBackend:
    """Auto-detect container runtime: prefer podman, fallback to docker."""
    if shutil.which("podman"):
        return ExecutionBackend.PODMAN
    if shutil.which("docker"):
        return ExecutionBackend.DOCKER
    raise RuntimeError("Neither podman nor docker found")


@dataclass
class CLIResult:
    stdout: str
    stderr: str
    code: int


class CLIStartError(RuntimeError):
    """Raised when a CLI fails to spawn."""


class CLIExecutionError(RuntimeError):
    """Raised when a CLI exits with a non-zero status."""

    def __init__(
        self,
        message: str,
        *,
        code: int,
        stdout: str,
        stderr: str,
        cmd: list[str],
    ) -> None:
        super().__init__(message)
        self.code = code
        self.stdout = stdout
        self.stderr = stderr
        self.cmd = cmd

    def __str__(self) -> str:
        parts = [super().__str__()]
        if self.stderr:
            parts.append(f"STDERR:\n{self.stderr.rstrip()}")
        if self.stdout:
            parts.append(f"STDOUT:\n{self.stdout.rstrip()}")
        return "\n".join(parts)


class BaseCLIConfig(BaseModel):
    """Common config fields shared by CLI wrappers."""

    mcp_host: str
    mcp_port: int
    provider_url: str
    image: str
    image_workdir: str = "/work"
    env: Optional[dict[str, str]] = None


class ProcessRunner:
    """Shared subprocess execution logic (spawn + communicate + timeout + kill).

    Concrete runners implement how cmd/env/cfg_root are transformed.
    """

    def __init__(self, log_path, logger):
        super().__init__()
        self.log_path = log_path
        self._logger = logger
        self._last_cmd: list[str] = []
        self._proc: Optional[subprocess.Popen[str]] = None

    def _write_log(
        self,
        stdout: str,
        stderr: str,
        *,
        returncode: int,
        config: BaseCLIConfig,
    ) -> None:
        if not self.log_path:
            return
        self.log_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            config_json = config.model_dump_json(indent=2)
        except Exception:
            config_json = str(config)
        with open(self.log_path, "w", encoding="utf-8") as fh:
            fh.write(f"Command: {shlex.join(self._last_cmd)}\n")
            fh.write("Config:\n")
            fh.write(f"{config_json}\n")
            fh.write(f"Exit code: {returncode}\n\n")
            if stdout:
                fh.write("STDOUT:\n")
                fh.write(stdout)
                if not stdout.endswith("\n"):
                    fh.write("\n")
            if stderr:
                if stdout:
                    fh.write("\n")
                fh.write("STDERR:\n")
                fh.write(stderr)
                if not stderr.endswith("\n"):
                    fh.write("\n")

    def run(
        self,
        *,
        cmd: list[str],
        env: dict[str, str],
        cfg_root: Path,
        config: BaseCLIConfig,
        spawn_error_message: str,
        stdin_devnull: bool = False,
    ) -> CLIResult:
        self._last_cmd = cmd
        stdout: str = ""
        stderr: str = ""
        code: int = -1

        popen_stdin = subprocess.DEVNULL if stdin_devnull else None
        timeout_s: Optional[float] = None

        try:
            self._proc = subprocess.Popen(
                cmd,
                stdin=popen_stdin,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                env=env,
            )
        except Exception as exc:
            stderr = f"{spawn_error_message}: {exc}"
            self._write_log(stdout, stderr, returncode=code, config=config)
            raise CLIStartError(f"{spawn_error_message}: {exc}") from exc

        try:
            try:
                stdout, stderr = self._proc.communicate(timeout=timeout_s)
            except subprocess.TimeoutExpired:
                self._logger.warning("CLI timed out; terminating: %s", shlex.join(cmd))
                self._proc.terminate()
                try:
                    stdout, stderr = self._proc.communicate(timeout=5)
                except subprocess.TimeoutExpired:
                    self._logger.warning("CLI did not terminate; killing it: %s", shlex.join(cmd))
                    self._proc.kill()
                    stdout, stderr = self._proc.communicate()
            code = self._proc.returncode or 0
        finally:
            self._write_log(stdout or "", stderr or "", returncode=code, config=config)
            if code != 0:
                self._logger.warning("CLI exited non-zero (%s): %s", code, shlex.join(cmd))
        if code != 0:
            raise CLIExecutionError(
                f"CLI exited non-zero ({code}): {shlex.join(cmd)}",
                code=code,
                stdout=stdout or "",
                stderr=stderr or "",
                cmd=cmd,
            )
        return CLIResult(stdout=stdout or "", stderr=stderr or "", code=code)

    def close(self) -> None:
        proc = self._proc
        if self._proc and self._proc.poll() is None:
            self._proc.terminate()
            try:
                self._proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                self._logger.warning("CLI process did not terminate; killing it.")
                self._proc.kill()
                try:
                    proc.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    self._logger.warning("CLI process did not exit after kill.")


class ContainerRunner(ProcessRunner):
    """Shared helpers for container-based runners."""

    host_gateway: str = "host.docker.internal"

    def _rewrite_mcp_config_path(self, inner_cmd: list[str], workdir: str) -> list[str]:
        if "--mcp-config" in inner_cmd:
            i = inner_cmd.index("--mcp-config")
            if i + 1 < len(inner_cmd):
                inner_cmd[i + 1] = f"{workdir}/mcp.json"
        return inner_cmd

    def _container_env_from(
        self,
        env: dict[str, str],
        host_gateway: str,
        config: BaseCLIConfig,
    ) -> dict[str, str]:
        forwarded: dict[str, str] = {}

        for k in (
            # Anthropic (Claude Code)
            "ANTHROPIC_BASE_URL",
            "ANTHROPIC_API_KEY",
            "ANTHROPIC_AUTH_TOKEN",
            # OpenAI (Codex)
            "OPENAI_API_KEY",
            "OPENAI_API_BASE",
            # Google (Gemini)
            "GEMINI_API_KEY",
            "GOOGLE_GEMINI_BASE_URL",
        ):
            if k in env:
                forwarded[k] = env[k]
        for k, v in env.items():
            if k.startswith("EXGENTIC_CTX_"):
                forwarded[k] = v
        if config.env:
            for k, v in config.env.items():
                forwarded[k] = v

        for k in (
            "HTTP_PROXY",
            "HTTPS_PROXY",
            "NO_PROXY",
            "http_proxy",
            "https_proxy",
            "no_proxy",
        ):
            if k in env:
                forwarded[k] = env[k]

        if "PYTHONIOENCODING" in env:
            forwarded["PYTHONIOENCODING"] = env["PYTHONIOENCODING"]

        for k, v in forwarded.items():
            if "://127.0.0.1:" in v:
                forwarded[k] = v.replace("://127.0.0.1:", f"://{host_gateway}:")
            if "://localhost:" in v:
                forwarded[k] = v.replace("://localhost:", f"://{host_gateway}:")

        return forwarded

    def _patch_mcp_json(self, *, cfg_root: Path, host_gateway: str) -> None:
        mcp_path = cfg_root / "mcp.json"
        if not mcp_path.exists():
            return

        try:
            data = json.loads(mcp_path.read_text(encoding="utf-8"))
            env_cfg = data.get("mcpServers", {}).get("environment", {})
            url = env_cfg.get("url", "")
            if not isinstance(url, str) or not url:
                return

            if "://127.0.0.1:" in url:
                data["mcpServers"]["environment"]["url"] = url.replace("://127.0.0.1:", f"://{host_gateway}:")
                mcp_path.write_text(json.dumps(data, indent=2), encoding="utf-8")
                self._logger.debug("Patched mcp.json for container host gateway: %s", mcp_path)
            elif "://localhost:" in url:
                data["mcpServers"]["environment"]["url"] = url.replace("://localhost:", f"://{host_gateway}:")
                mcp_path.write_text(json.dumps(data, indent=2), encoding="utf-8")
                self._logger.debug("Patched mcp.json for container host gateway: %s", mcp_path)

        except Exception as exc:
            self._logger.warning("Failed to patch mcp.json (%s): %s", mcp_path, exc)


class PodmanRunner(ContainerRunner):
    """Run the inner command inside a container via Podman."""

    host_gateway = "host.containers.internal"

    def __init__(self, log_path, logger):
        super().__init__(log_path, logger)

    def run(
        self,
        *,
        cmd: list[str],
        env: dict[str, str],
        cfg_root: Path,
        config: BaseCLIConfig,
        spawn_error_message: str,
    ) -> CLIResult:
        runtime = "podman"
        host_gateway = "host.containers.internal"
        host_cfg_root = str(cfg_root.resolve())

        # Patch mcp.json so container uses host gateway (not 127.0.0.1/localhost)
        self._patch_mcp_json(cfg_root=cfg_root, host_gateway=host_gateway)

        inner_cmd = self._rewrite_mcp_config_path(list(cmd), workdir=config.image_workdir)

        # Minimal env forwarding into container (encoded via podman -e flags)
        container_env = self._container_env_from(env, host_gateway=host_gateway, config=config)
        container_env["HOME"] = config.image_workdir
        connection_args = self._resolve_podman_connection_args()
        user_args: list[str] = []
        uid = getattr(os, "getuid", None)
        gid = getattr(os, "getgid", None)
        if callable(uid) and callable(gid):
            try:
                user_args = ["--user", f"{uid()}:{gid()}"]
            except Exception:
                user_args = []
        wrapped_cmd: list[str] = [
            runtime,
            *connection_args,
            "run",
            "--rm",
            *user_args,
            "-v",
            f"{host_cfg_root}:{config.image_workdir}:Z",
            "-w",
            config.image_workdir,
        ]
        for k, v in container_env.items():
            wrapped_cmd.extend(["-e", f"{k}={v}"])
        wrapped_cmd.append(str(config.image))
        wrapped_cmd.extend(inner_cmd)

        # Important: do NOT keep stdin open (avoids "podman run never ends")
        return super().run(
            cmd=wrapped_cmd,
            env=env,
            cfg_root=cfg_root,
            config=config,
            spawn_error_message=spawn_error_message,
            stdin_devnull=True,
        )

    def _resolve_podman_connection_args(self) -> list[str]:
        """Return extra args for the `podman` CLI to select a connection.

        Priority:
        1) PODMAN_CONNECTION env var
        2) auto-detect default connection from `podman system connection list --format json`
        3) fallback: no args (let Podman decide; works on native Linux / preconfigured env)
        """
        # 1) environment override (nice for CI/users)
        env_name = os.environ.get("PODMAN_CONNECTION")
        if env_name:
            return ["--connection", env_name]

        # 2) auto-detect default connection
        try:
            proc = subprocess.run(
                ["podman", "system", "connection", "list", "--format", "json"],
                check=False,
                capture_output=True,
                text=True,
            )
            if proc.returncode != 0:
                self._logger.info(
                    "podman system connection list failed (rc=%s): %s",
                    proc.returncode,
                    (proc.stderr or "").strip(),
                )
                return []

            data = json.loads(proc.stdout or "[]")
            # entries look like: {"Name": "...", "URI": "...", "Identity": "...", "Default": true, ...}
            default = next((x for x in data if x.get("Default") is True), None)
            if default and default.get("Name"):
                return [
                    "--url",
                    str(default["URI"]),
                    "--identity",
                    str(default["Identity"]),
                ]
        except Exception as exc:
            self._logger.info("Failed to auto-detect Podman connection: %r", exc)
            return []

        return []


class DockerRunner(ContainerRunner):
    """Run the inner command inside a container via Docker."""

    def __init__(self, log_path, logger):
        super().__init__(log_path, logger)

    def run(
        self,
        *,
        cmd: list[str],
        env: dict[str, str],
        cfg_root: Path,
        config: BaseCLIConfig,
        spawn_error_message: str,
    ) -> CLIResult:
        runtime = "docker"
        host_gateway = "host.docker.internal"
        host_cfg_root = str(cfg_root.resolve())

        # Patch mcp.json so container uses host gateway (not 127.0.0.1/localhost)
        self._patch_mcp_json(cfg_root=cfg_root, host_gateway=host_gateway)

        inner_cmd = self._rewrite_mcp_config_path(list(cmd), workdir=config.image_workdir)

        # Minimal env forwarding into container (encoded via docker -e flags)
        container_env = self._container_env_from(env, host_gateway=host_gateway, config=config)
        container_env["HOME"] = config.image_workdir
        user_args: list[str] = []
        uid = getattr(os, "getuid", None)
        gid = getattr(os, "getgid", None)
        if callable(uid) and callable(gid):
            try:
                user_args = ["--user", f"{uid()}:{gid()}"]
            except Exception:
                user_args = []
        wrapped_cmd: list[str] = [
            runtime,
            "run",
            "--rm",
            "--add-host",
            f"{host_gateway}:host-gateway",
            *user_args,
            "-v",
            f"{host_cfg_root}:{config.image_workdir}",
            "-w",
            config.image_workdir,
        ]
        for k, v in container_env.items():
            wrapped_cmd.extend(["-e", f"{k}={v}"])
        wrapped_cmd.append(str(config.image))
        wrapped_cmd.extend(inner_cmd)

        # Important: do NOT keep stdin open (avoids "docker run never ends")
        return super().run(
            cmd=wrapped_cmd,
            env=env,
            cfg_root=cfg_root,
            config=config,
            spawn_error_message=spawn_error_message,
            stdin_devnull=True,
        )



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from typing import Any, ClassVar

from ...core.agent import Agent
from ...core.types import ModelSettings


class SmolagentBaseAgent(Agent):
    display_name: ClassVar[str] = "SmolAgents Base Agent"
    slug_name: ClassVar[str] = "smolagents_base"

    model: str = "watsonx/meta-llama/llama-3-3-70b-instruct"
    max_steps: int = 150
    model_settings: ModelSettings | None = None
    retry_on_all_errors: bool = True

    def get_instance_kwargs(
        self,
        session_id: str,
    ) -> dict[str, Any]:
        return {
            "session_id": session_id,
            "model_id": self.model,
            "max_steps": self.max_steps,
            "model_settings": self.model_settings,
            "retry_on_all_errors": self.retry_on_all_errors,
        }

    @property
    def model_name(self) -> str:  # type: ignore[override]
        return str(self.model).split("/")[-1]

    def get_models_names(self) -> list[str]:  # type: ignore[override]
        return [str(self.model)]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""HTTPTransport + serve() — run any object as an HTTP service."""

from __future__ import annotations

import base64
import threading
import time
from typing import Any, Optional

import cloudpickle as cp
import httpx
from fastapi import FastAPI
from pydantic import BaseModel as PydanticBaseModel

from .transport import ObjectHost, ObjectProxy, Transport, deserialize_error, serialize_error

# ── HTTP models ──────────────────────────────────────────────────────


class CallRequest(PydanticBaseModel):
    method: str
    args: str  # base64(cloudpickle)
    kwargs: str  # base64(cloudpickle)


class GetRequest(PydanticBaseModel):
    name: str


class SetRequest(PydanticBaseModel):
    name: str
    value: str  # base64(cloudpickle)


class RPCResponse(PydanticBaseModel):
    status: str  # "ok" | "error"
    result: Optional[str] = None  # base64(cloudpickle)
    error_type: Optional[str] = None
    error_msg: Optional[str] = None
    error_tb: Optional[str] = None
    error_pickled: Optional[str] = None  # base64(cloudpickle'd exception)


# ── helpers ──────────────────────────────────────────────────────────


def _encode(obj: Any) -> str:
    return base64.b64encode(cp.dumps(obj)).decode("ascii")


def _decode(data: str) -> Any:
    return cp.loads(base64.b64decode(data))


def _error_response(exc: Exception) -> RPCResponse:
    data = serialize_error(exc)
    pickled_b64 = None
    if data["pickled"] is not None:
        pickled_b64 = base64.b64encode(data["pickled"]).decode("ascii")
    return RPCResponse(
        status="error",
        error_type=data["type"],
        error_msg=data["msg"],
        error_tb=data["tb"],
        error_pickled=pickled_b64,
    )


# ── FastAPI app ──────────────────────────────────────────────────────


def create_app(host: ObjectHost) -> FastAPI:
    app = FastAPI()

    @app.get("/health")
    def health():
        return {"status": "ok"}

    @app.post("/call")
    def handle_call(req: CallRequest) -> RPCResponse:
        try:
            result = host.handle("call", req.method, *_decode(req.args), **_decode(req.kwargs))
            return RPCResponse(status="ok", result=_encode(result))
        except Exception as exc:
            return _error_response(exc)

    @app.post("/get")
    def handle_get(req: GetRequest) -> RPCResponse:
        try:
            return RPCResponse(status="ok", result=_encode(host.handle("get", req.name)))
        except Exception as exc:
            return _error_response(exc)

    @app.post("/set")
    def handle_set(req: SetRequest) -> RPCResponse:
        try:
            host.handle("set", req.name, _decode(req.value))
            return RPCResponse(status="ok")
        except Exception as exc:
            return _error_response(exc)

    return app


# ── serve() ──────────────────────────────────────────────────────────


def serve(obj: Any, host: str = "0.0.0.0", port: int = 8080) -> None:
    """Serve an object over HTTP (blocking)."""
    import uvicorn

    uvicorn.run(create_app(ObjectHost(obj)), host=host, port=port, log_level="warning")


# ── HTTPTransport — client side ──────────────────────────────────────


class HTTPTransport(Transport):
    """Talks to an HTTP server hosting an ObjectHost."""

    def __init__(self, base_url: str, timeout: float = 30.0) -> None:
        self._base_url = base_url.rstrip("/")
        self._client = httpx.Client(timeout=timeout)

    def _rpc(self, endpoint: str, payload: dict) -> Any:
        resp = self._client.post(f"{self._base_url}{endpoint}", json=payload)
        resp.raise_for_status()
        data = RPCResponse(**resp.json())
        if data.status == "error":
            pickled = base64.b64decode(data.error_pickled) if data.error_pickled else None
            raise deserialize_error(
                {
                    "type": data.error_type or "RuntimeError",
                    "msg": data.error_msg or "",
                    "tb": data.error_tb or "",
                    "pickled": pickled,
                }
            )
        return _decode(data.result) if data.result is not None else None

    def call(self, method: str, *args: Any, **kwargs: Any) -> Any:
        return self._rpc(
            "/call",
            {
                "method": method,
                "args": _encode(args),
                "kwargs": _encode(kwargs),
            },
        )

    def get(self, name: str) -> Any:
        return self._rpc("/get", {"name": name})

    def set(self, name: str, value: Any) -> None:
        self._rpc("/set", {"name": name, "value": _encode(value)})

    def close(self) -> None:
        self._client.close()

    def __repr__(self) -> str:
        return f"HTTPTransport({self._base_url!r})"


# ── Utilities ────────────────────────────────────────────────────────


def _wait_for_health(url: str, timeout: float = 15.0) -> None:
    deadline = time.monotonic() + timeout
    while time.monotonic() < deadline:
        try:
            if httpx.get(f"{url}/health", timeout=2.0).status_code == 200:
                return
        except httpx.HTTPError:
            pass
        time.sleep(0.1)
    raise TimeoutError(f"Service at {url} did not become healthy within {timeout}s")


# ── ServiceRunner ────────────────────────────────────────────────────


class ServiceRunner:
    """Starts an HTTP service in a background thread and returns an ObjectProxy."""

    def __init__(
        self,
        target_cls: type,
        *args: Any,
        port: int | None = None,
        **kwargs: Any,
    ) -> None:
        self._target_cls = target_cls
        self._args = args
        self._kwargs = kwargs
        from ._utils import find_free_port

        self._port = port or find_free_port()
        self._server = None

    def start(self) -> ObjectProxy:
        import uvicorn

        from ...core.context import set_context_fallback, try_get_context

        # Set process-wide fallback so context is available in uvicorn's
        # request handler threads (which don't inherit ContextVar).
        set_context_fallback(try_get_context())

        obj = self._target_cls(*self._args, **self._kwargs)
        app = create_app(ObjectHost(obj))

        config = uvicorn.Config(app, host="127.0.0.1", port=self._port, log_level="warning")
        self._server = uvicorn.Server(config)
        threading.Thread(target=self._server.run, daemon=True).start()

        url = f"http://127.0.0.1:{self._port}"
        _wait_for_health(url)

        transport = HTTPTransport(url)
        proxy = ObjectProxy(transport)

        server_ref = self._server

        def _close() -> None:
            try:
                transport.call("close")
            except AttributeError:
                pass
            transport.close()
            server_ref.should_exit = True
            set_context_fallback(None)

        object.__setattr__(proxy, "close", _close)
        return proxy



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""BFCL benchmark adapter."""

from .bfcl_benchmark import BFCLBenchmark

__all__ = ["BFCLBenchmark"]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import json

from ...utils.paths import RunPaths

FUNNEL = [
    ("edit", lambda s: (s.get("num_edit_commands") or 0) > 0),
    ("call_submit", lambda s: s.get("agent_call_submit") == 1),
    ("non_empty_patch", lambda s: s.get("patch_non_empty") == 1),
    ("container", lambda s: s.get("container_status") == 1),
    ("fail_to_pass", lambda s: s.get("fail_to_pass_rate") == 100.0),
    ("pass_to_pass", lambda s: s.get("pass_to_pass_rate") == 100.0),
    ("score", lambda s: s.get("score") == 1),
]


def collect_metrics(run_id, session_ids: list[str]) -> dict:
    from ...core.context import get_context

    run_paths = RunPaths(run_id=run_id, output_dir=get_context().output_dir)
    summaries = []
    for sid in session_ids:
        path = run_paths.session(sid).benchmark_results
        if path.exists():
            summaries.append(json.loads(path.read_text()).get("summary", {}))

    return {
        "funnel": {"total": len(summaries)} | {name: sum(1 for s in summaries if check(s)) for name, check in FUNNEL}
    }



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from .base_agent import SmolagentBaseAgent
from .code_agent import SmolagentCodeAgent
from .tool_calling_agent import SmolagentToolCallingAgent

__all__ = [
    "SmolagentBaseAgent",
    "SmolagentCodeAgent",
    "SmolagentToolCallingAgent",
]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Evaluator ABC — benchmark evaluation logic (task discovery, session config, aggregation).

An Evaluator can be wrapped with ``with_runner()`` for container isolation,
keeping heavy dependencies off the host.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from ..utils.paths import SessionPaths, get_run_paths
from .types import BenchmarkResults, SessionIndex


class Evaluator(ABC):
    """Benchmark evaluation logic — task discovery, session config, aggregation.

    Runs in the same isolation level as the benchmark's runner (can be containerized).
    Returns only simple serializable data across the transport boundary.
    """

    @abstractmethod
    def list_tasks(self) -> list[str]:
        """Return available task identifiers for this benchmark."""
        ...

    @abstractmethod
    def get_session_kwargs(self, index: SessionIndex) -> dict[str, Any]:
        """Return kwargs for constructing the Session for a given task.

        The orchestrator will call::

            with_runner(session_class, runner=..., **session_kwargs, **runner_kwargs)
        """
        ...

    @abstractmethod
    def aggregate_sessions(self, sessions: list[SessionIndex]) -> BenchmarkResults:
        """Aggregate results for the specified task sessions."""
        ...

    def get_sessions_paths(self, sessions: list[SessionIndex]) -> list[SessionPaths]:
        """Return ``SessionPaths`` for each session index."""
        run_paths = get_run_paths()
        return [run_paths.session(s.session_id) for s in sessions]

    def close(self) -> None:
        """Optional cleanup hook."""
        return



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import os
import sys
from pathlib import Path

import pytest
import requests
from exgentic.integrations.litellm import LitellmProxy


@pytest.mark.skipif(
    not (Path(sys.executable).parent / "litellm").exists(),
    reason="litellm CLI not installed in active venv",
)
def test_proxy_reuses_cached_completion_response(tmp_path, fake_openai_server) -> None:
    fake_openai_server.RequestHandlerClass.request_count = 0
    backend_port = fake_openai_server.server_address[1]
    backend_base = f"http://127.0.0.1:{backend_port}/v1"

    venv_bin = Path(sys.executable).parent
    env = {
        "OPENAI_API_BASE": backend_base,
        "OPENAI_API_KEY": "test-key",  # pragma: allowlist secret
        "EXGENTIC_CACHE_DIR": str(tmp_path / "cache"),
        "EXGENTIC_litellm_cache_dir": ".litellm_cache",
        "EXGENTIC_litellm_caching": "true",
        "EXGENTIC_LOG_LEVEL": "DEBUG",
        "PATH": f"{venv_bin}{os.pathsep}{os.environ.get('PATH', '')}",
    }
    payload = {
        "model": "openai/gpt-4o-mini",
        "messages": [{"role": "user", "content": "Hello"}],
        "temperature": 0.0,
    }

    with LitellmProxy(
        model="openai/gpt-4o-mini",
        env=env,
        startup_timeout=10.0,
    ) as proxy:
        url = f"{proxy.base_url}/v1/chat/completions"
        first = requests.post(url, json=payload, timeout=10)
        first.raise_for_status()
        second = requests.post(url, json=payload, timeout=10)
        second.raise_for_status()

    assert fake_openai_server.RequestHandlerClass.request_count == 1



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

from typing import Any, Literal

from pydantic import create_model

from ...core.types import ActionType, SingleAction
from .json_schema import make_args_model_from_json_schema


def openai_tools_to_action_types(tools: list[dict[str, Any]]) -> list[ActionType]:
    """Translate OpenAI-style tools into ActionType definitions.

    Builds concrete argument models from each tool's parameter schema so no
    information is lost when emitting tools back to the LLM via ActionType.
    """
    actions: list[ActionType] = []
    for t in tools:
        if not isinstance(t, dict) or t.get("type") != "function":
            continue
        fn = t.get("function") or {}
        name = fn.get("name")
        if not isinstance(name, str):
            continue
        desc = fn.get("description") or ""
        params = fn.get("parameters") or {}

        args_model = make_args_model_from_json_schema(name, params)

        action_model = create_model(
            f"{name}_Action",
            __base__=SingleAction,
            name=(Literal[name], name),
            arguments=(args_model, ...),
        )
        actions.append(ActionType(name=name, description=str(desc), cls=action_model))

    if not actions:
        raise ValueError("No OpenAI function tools provided to translate into ActionTypes")
    return actions


def mcp_to_openai_tool(mcp_tool: Any) -> dict[str, Any]:
    """Converts a tool definition from a 'mcp' format  into  OpenAI tool schema."""
    function_name = mcp_tool.name
    description = mcp_tool.description or ""
    parameters_schema = mcp_tool.inputSchema or {"type": "object", "properties": {}}

    if not function_name:
        raise ValueError("MCP tool definition is missing a 'name'.")

    tool_schema = {
        "type": "function",
        "function": {
            "name": function_name,
            "description": description,
            "parameters": parameters_schema,
        },
    }
    return tool_schema


def mcp_tools_to_openai_tools(mcp_tools: list[dict[str, Any]]) -> list[dict[str, Any]]:
    return [mcp_to_openai_tool(mcp_tool) for mcp_tool in mcp_tools]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import json
import threading
from typing import Any, Optional

from ...core.orchestrator.observer import Observer


class ConfigsObserver(Observer):
    def __init__(self, run_id: str | None = None) -> None:
        super().__init__(run_id)
        self._lock = threading.Lock()
        self._run_config: Optional[Any] = None

    def on_run_start(self, run_config) -> None:
        with self._lock:
            self._run_config = run_config
        self._write_config(run_config)

    def on_run_success(self, results, run_config) -> None:
        with self._lock:
            self._run_config = run_config
        self._write_config(run_config)

    def on_run_error(self, error) -> None:
        with self._lock:
            run_config = self._run_config
        if run_config is None:
            return
        self._write_config(run_config)

    def _write_config(self, run_config) -> None:
        rp = self.paths
        config_path = rp.config
        try:
            config_path.parent.mkdir(parents=True, exist_ok=True)
            with open(config_path, "w", encoding="utf-8") as f:
                json.dump(
                    run_config.model_dump(mode="json"),
                    f,
                    ensure_ascii=False,
                    indent=2,
                )
        except OSError:
            # Read-only runs should still be able to aggregate without persisting.
            return



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import json
from typing import Any, ClassVar

from ..core.benchmark import Benchmark
from ..core.evaluator import Evaluator
from ..core.session import Session
from ..core.types import (
    Action,
    ActionType,
    BenchmarkResults,
    SessionIndex,
    SessionScore,
    SingleObservation,
)
from .agent import (
    BAD_ACTION_TYPE,
    FINISH_ACTION_TYPE,
    GOOD_ACTION_TYPE,
    BadAction,
    FinishAction,
    GoodAction,
)


class TestSession(Session):
    __test__ = False

    def __init__(
        self,
        *,
        task_id: str,
        session_id: str,
        stop_on_step: bool,
        invalid_observation: bool,
    ) -> None:
        self._session_id = session_id
        self._task_id = task_id
        self._stop_on_step = stop_on_step
        self._invalid_observation = invalid_observation
        self._done = False
        self._good = 0
        self._bad = 0
        self._steps = 0
        super().__init__()

    @property
    def task_id(self) -> str:
        return self._task_id

    @property
    def task(self) -> str:
        return f"Task {self._task_id}"

    @property
    def context(self) -> dict[str, Any]:
        return {"task_id": self._task_id}

    @property
    def actions(self) -> list[ActionType]:
        return [GOOD_ACTION_TYPE, BAD_ACTION_TYPE, FINISH_ACTION_TYPE]

    def start(self) -> SingleObservation:
        return SingleObservation(result="start")

    def step(self, action: Action) -> SingleObservation | None:
        self._steps += 1
        if self._invalid_observation:
            return "invalid-observation"  # type: ignore[return-value]
        if isinstance(action, GoodAction):
            self._good += 1
        elif isinstance(action, BadAction):
            self._bad += 1
        elif isinstance(action, FinishAction):
            self._done = True
            return SingleObservation(result="finish")
        if self._stop_on_step:
            return None
        return SingleObservation(result="step")

    def done(self) -> bool:
        return self._done

    def score(self) -> SessionScore:
        total = self._good + self._bad
        score = float(self._good / total) if total > 0 else 0.0
        success = bool(self._done and self._bad == 0 and total > 0)
        result = SessionScore(
            score=score,
            success=success,
            is_finished=self._done,
            session_metrics={"good": self._good, "bad": self._bad, "total": total},
            session_metadata={"steps": self._steps},
        )
        self.save_standard_results(result)
        return result

    def get_config(self) -> dict[str, Any]:
        return {
            "task_id": self._task_id,
            "stop_on_step": self._stop_on_step,
        }

    def close(self) -> None:
        return None


class TestEvaluator(Evaluator):
    __test__ = False

    def __init__(
        self,
        *,
        tasks: list[str] | None = None,
        stop_on_step: bool = False,
        invalid_observation: bool = False,
    ) -> None:
        self._tasks = tasks or ["task-1", "task-2", "task-3"]
        self._stop_on_step = stop_on_step
        self._invalid_observation = invalid_observation

    def list_tasks(self) -> list[str]:
        return list(self._tasks)

    def get_session_kwargs(self, index: SessionIndex) -> dict[str, Any]:
        return {
            "task_id": str(index.task_id),
            "session_id": index.session_id,
            "stop_on_step": self._stop_on_step,
            "invalid_observation": self._invalid_observation,
        }

    def aggregate_sessions(self, sessions: list[SessionIndex]) -> BenchmarkResults:
        scores: list[float] = []
        for paths in self.get_sessions_paths(sessions):
            if not paths.results.exists():
                continue
            payload = json.loads(paths.results.read_text(encoding="utf-8"))
            try:
                score = float(payload["score"])
            except Exception:
                continue
            scores.append(score)
        avg = sum(scores) / len(scores) if scores else 0.0
        return BenchmarkResults(
            benchmark_name="test_benchmark",
            total_tasks=len(sessions),
            score=avg,
            metrics={},
        )


class TestBenchmark(Benchmark):
    __test__ = False
    display_name: ClassVar[str] = "Test Benchmark"
    slug_name: ClassVar[str] = "test_benchmark"
    runner: str | None = "direct"  # Run in-process; override via benchmark_kwargs for runner tests
    tasks: list[str] = ["task-1", "task-2", "task-3"]  # noqa: RUF012

    @classmethod
    def get_evaluator_class(cls):
        return TestEvaluator

    @classmethod
    def get_session_class(cls):
        return TestSession

    stop_on_step: bool = False
    invalid_observation: bool = False

    def get_evaluator_kwargs(self) -> dict[str, Any]:
        return {
            "tasks": self.tasks,
            "stop_on_step": self.stop_on_step,
            "invalid_observation": self.invalid_observation,
        }



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

from enum import Enum

from pydantic import BaseModel, field_validator


class RetryStrategy(str, Enum):
    EXPONENTIAL_BACKOFF = "exponential_backoff_retry"
    CONSTANT = "constant_retry"


class ModelSettings(BaseModel):
    temperature: float | None = 1.0
    top_p: float | None = None
    max_tokens: int | None = None
    reasoning_effort: str | None = None
    num_retries: int | None = 5
    retry_after: float = 0.5
    retry_strategy: RetryStrategy = RetryStrategy.EXPONENTIAL_BACKOFF

    @field_validator("temperature")
    @classmethod
    def _validate_temperature(cls, value: float | None) -> float | None:
        if value is not None and value < 0:
            raise ValueError("temperature must be >= 0")
        return value

    @field_validator("max_tokens")
    @classmethod
    def _validate_max_tokens(cls, value: int | None) -> int | None:
        if value is not None and value < 0:
            raise ValueError("max_tokens must be >= 0")
        return value

    @field_validator("top_p")
    @classmethod
    def _validate_top_p(cls, value: float | None) -> float | None:
        if value is None:
            return value
        if value < 0 or value > 1:
            raise ValueError("top_p must be between 0 and 1")
        return value

    @field_validator("num_retries")
    @classmethod
    def _validate_num_retries(cls, value: int | None) -> int | None:
        if value is None:
            return None
        if value < 0:
            raise ValueError("num_retries must be >= 0")
        return value

    @field_validator("retry_after")
    @classmethod
    def _validate_retry_after(cls, value: float) -> float:
        if value < 0:
            raise ValueError("retry_after must be >= 0")
        return value



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Integration test that simulates ``uv tool install exgentic``.

The test creates an isolated venv (the way ``uv tool install`` would),
installs exgentic into it, then runs ``exgentic setup --benchmark <slug>``
from a clean working directory that has **no** ``pyproject.toml`` in any
parent — exactly the situation a user hits after a global tool install.
"""

from __future__ import annotations

import shutil
import subprocess
import sys
from pathlib import Path

import pytest

_uv_available = shutil.which("uv") is not None

# Root of the exgentic source tree (two levels up from this file).
_REPO_ROOT = Path(__file__).resolve().parents[2]

# Per-benchmark setup timeout in seconds (5 minutes).
_SETUP_TIMEOUT = 300


_ALL_BENCHMARKS = [
    "tau2",
    "gsm8k",
    "appworld",
    "bfcl",
    "browsecompplus",
    "hotpotqa",
    "swebench",
]


@pytest.mark.skipif(not _uv_available, reason="uv CLI not available")
@pytest.mark.parametrize("benchmark", _ALL_BENCHMARKS)
def test_setup_in_tool_install_venv(benchmark: str, tmp_path: Path) -> None:
    """Install exgentic into a fresh venv, then run ``exgentic setup``."""
    venv_dir = tmp_path / "venv"
    work_dir = tmp_path / "workdir"
    work_dir.mkdir()

    # 1. Create an isolated venv using uv.
    subprocess.run(
        ["uv", "venv", str(venv_dir), "--python", f"{sys.version_info.major}.{sys.version_info.minor}"],
        check=True,
        capture_output=True,
        timeout=60,
    )

    # 2. Install exgentic from the local source tree into the venv.
    subprocess.run(
        ["uv", "pip", "install", str(_REPO_ROOT), "--python", str(venv_dir / "bin" / "python")],
        check=True,
        capture_output=True,
        timeout=300,
    )

    # 3. Locate the ``exgentic`` entry-point inside the venv.
    exgentic_bin = venv_dir / "bin" / "exgentic"
    assert exgentic_bin.exists(), f"exgentic CLI not found at {exgentic_bin}"

    # 4. Run ``exgentic setup --benchmark <slug>`` from the clean workdir.
    #    The working directory deliberately has no pyproject.toml ancestors.
    result = subprocess.run(
        [str(exgentic_bin), "setup", "--benchmark", benchmark, "--force"],
        cwd=str(work_dir),
        capture_output=True,
        text=True,
        timeout=_SETUP_TIMEOUT,
    )

    assert result.returncode == 0, (
        f"exgentic setup --benchmark {benchmark} failed "
        f"(rc={result.returncode}).\nstdout:\n{result.stdout}\nstderr:\n{result.stderr}"
    )

    # 5. Verify the installation marker was written.
    venv_python = str(venv_dir / "bin" / "python")
    check = subprocess.run(
        [
            venv_python,
            "-c",
            (
                "from exgentic.utils.installation_tracker import is_installed; "
                f"assert is_installed({benchmark!r}, 'benchmark'), "
                f"'installation marker not found for {benchmark}'"
            ),
        ],
        cwd=str(work_dir),
        capture_output=True,
        text=True,
        timeout=30,
    )
    assert check.returncode == 0, (
        f"Installation marker check failed for {benchmark}.\n" f"stdout:\n{check.stdout}\nstderr:\n{check.stderr}"
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import logging
import os
import threading
from typing import Any

from nicegui import ui

from ....core.context import run_scope
from ....core.types import RunConfig
from ....observers.handlers.dashboard_events import DashboardEventsObserver
from ....utils.paths import RunPaths
from ....utils.settings import get_settings
from ...lib.api import (
    evaluate,
    get_agent_info,
    get_benchmark_info,
    load_agent_class,
    load_benchmark_class,
)
from .data import (
    _build_history_sessions,
    _load_history_turns,
    _load_run_context,
    _open_session_from_row,
    _resolve_tab_label,
    get_display_mappings,
    load_leaderboard_data,
)
from .forms import _build_agent_form, _build_pydantic_form, _collect_values
from .panels import (
    _render_dive_panel,
    _render_overview_panel,
    _render_run_log,
)
from .state import LEADERBOARD_COLUMNS, RunState, RunViews
from .status import _status_from_outcome

_LOG = logging.getLogger("exgentic.interfaces.dashboard")
if not _LOG.handlers:
    _handler = logging.StreamHandler()
    _handler.setFormatter(logging.Formatter("Dashboard | %(message)s"))
    _LOG.addHandler(_handler)
_LOG.setLevel(logging.INFO)
_LOG.propagate = False


def _init_tabs(
    state: RunState,
    scope: str,
    labels: list[str],
) -> tuple[Any, dict[str, Any], Any]:
    if scope not in state.active_tabs:
        state.active_tabs[scope] = labels[0]

    with ui.tabs() as tabs:
        tab_by_name = {label: ui.tab(label) for label in labels}

    state.tabs_controls[scope] = tabs
    state.tabs_by_scope[scope] = tab_by_name

    def _on_tab_change(e) -> None:
        value = getattr(e, "value", None)
        state.active_tabs[scope] = _resolve_tab_label(value, tab_by_name, labels[0])

    tabs.on_value_change(_on_tab_change)

    active_label = state.active_tabs.get(scope, labels[0])
    active_tab = tab_by_name.get(active_label, tab_by_name[labels[0]])
    return tabs, tab_by_name, active_tab


def process_new_events(state: RunState) -> bool:
    tracker = state.tracker
    if tracker is None:
        return False

    events_to_process = []
    event_count = 0
    max_events = 40

    while not tracker.events.empty() and event_count < max_events:
        try:
            evt = tracker.events.get_nowait()
            events_to_process.append(evt)
            event_count += 1
        except Exception:
            break

    if not events_to_process:
        return False

    sessions = state.sessions
    turns = state.turns

    for evt in events_to_process:
        state.events.append(evt)
        et = evt.get("type")

        if et == "session_started":
            sid = evt.get("session_id")
            if sid:
                sessions[sid] = {
                    "status": "running",
                    "steps": 0,
                    "success": None,
                    "is_finished": None,
                    "error_source": None,
                }

        elif et == "step":
            sid = evt.get("session_id")
            if sid and sid in sessions:
                sessions[sid]["steps"] = sessions[sid].get("steps", 0) + 1
                if "execution_time" in evt:
                    sessions[sid]["execution_time"] = evt.get("execution_time")
                if "agent_cost" in evt:
                    sessions[sid]["agent_cost"] = evt.get("agent_cost")
                if "benchmark_cost" in evt:
                    sessions[sid]["benchmark_cost"] = evt.get("benchmark_cost")
                step_no = evt.get("n")
                act_obj = evt.get("action_obj")
                if act_obj is not None:
                    turns.setdefault(sid, []).append({"type": "action", "step": step_no, "content": act_obj})
                    turns[sid] = turns[sid][-200:]

        elif et == "observation":
            sid = evt.get("session_id")
            if sid and sid in sessions:
                step_no = evt.get("step")
                obs_obj = evt.get("observation")
                if "execution_time" in evt:
                    sessions[sid]["execution_time"] = evt.get("execution_time")
                if "agent_cost" in evt:
                    sessions[sid]["agent_cost"] = evt.get("agent_cost")
                if "benchmark_cost" in evt:
                    sessions[sid]["benchmark_cost"] = evt.get("benchmark_cost")
                if obs_obj is not None:
                    turns.setdefault(sid, []).append({"type": "observation", "step": step_no, "content": obs_obj})
                    turns[sid] = turns[sid][-200:]

        elif et == "session_finished":
            sid = evt.get("session_id")
            if sid and sid in sessions:
                success = evt.get("success", False)
                is_finished = evt.get("is_finished")
                error_source = evt.get("error_source")
                sessions[sid]["status"] = _status_from_outcome(success, is_finished, error_source)
                sessions[sid]["success"] = success
                sessions[sid]["score"] = evt.get("score")
                if "steps" in evt:
                    sessions[sid]["steps"] = evt.get("steps")
                if "execution_time" in evt:
                    sessions[sid]["execution_time"] = evt.get("execution_time")
                if "agent_cost" in evt:
                    sessions[sid]["agent_cost"] = evt.get("agent_cost")
                if "benchmark_cost" in evt:
                    sessions[sid]["benchmark_cost"] = evt.get("benchmark_cost")
                if "is_finished" in evt:
                    sessions[sid]["is_finished"] = is_finished
                if "error_source" in evt:
                    sessions[sid]["error_source"] = error_source
                if "details" in evt:
                    details = evt.get("details") or {}
                    sessions[sid]["error"] = details.get("error")
                if not success:
                    details = evt.get("details") or {}
                    turns.setdefault(sid, []).append(
                        {
                            "type": "error",
                            "step": sessions[sid].get("steps", 0),
                            "content": details or "Session ended with an error.",
                        }
                    )

        elif et == "saved":
            ctx_mgr = state.run_context_manager
            if ctx_mgr is not None:
                try:
                    ctx_mgr.__exit__(None, None, None)
                except Exception:
                    pass
                state.run_context_manager = None
            state.run_active = False
            state.refresh_needed = True
            _LOG.info("run saved; dashboard state updated")

        elif et == "run_meta":
            run_id = evt.get("run_id")
            if run_id:
                state.run_id = run_id
                _LOG.info("run_meta received: %s", run_id)

    return True


@ui.refreshable
def bench_form_panel(state: RunState) -> None:
    bench_key = state.bench_key
    if bench_key is None:
        ui.label("No benchmark selected")
        return
    try:
        bench_cls = load_benchmark_class(bench_key)
    except Exception as exc:
        ui.label(f"Failed to load benchmark '{bench_key}': {exc}")
        return
    state.bench_controls = _build_pydantic_form(bench_cls, disabled=state.run_active)


@ui.refreshable
def agent_form_panel(state: RunState) -> None:
    agent_key = state.agent_key
    if agent_key is None:
        ui.label("No agent selected")
        return
    try:
        agent_cls = load_agent_class(agent_key)
    except Exception as exc:
        ui.label(f"Failed to load agent '{agent_key}': {exc}")
        return
    state.agent_controls = _build_agent_form(agent_cls, disabled=state.run_active)


@ui.refreshable
def run_log_panel(state: RunState) -> None:
    _render_run_log(state.run_id)


@ui.refreshable
def overview_panel(state: RunState) -> None:
    fallback_benchmark = None
    if state.bench_key:
        try:
            fallback_benchmark = get_benchmark_info(state.bench_key)["display_name"]
        except Exception:
            fallback_benchmark = None
    fallback_agent = None
    if state.agent_key:
        try:
            fallback_agent = get_agent_info(state.agent_key)["display_name"]
        except Exception:
            fallback_agent = None
    context = _load_run_context(
        state.run_id,
        fallback_benchmark=fallback_benchmark,
        fallback_agent=fallback_agent,
        planned_fallback=state.planned_sessions,
        workers_fallback=state.max_workers if state.run_active else None,
    )

    def _open_session(row: dict) -> None:
        _open_session_from_row(state, "run", row, dive_panel.refresh)

    _render_overview_panel(
        state.sessions,
        run_active=state.run_active,
        run_id=state.run_id,
        results=context.results,
        planned_sessions=context.planned_sessions,
        total_workers=context.total_workers,
        run_meta=context.run_meta,
        on_open_session=_open_session,
    )


@ui.refreshable
def dive_panel(state: RunState) -> None:
    sessions = sorted(state.sessions.keys())
    if not sessions:
        ui.label("No sessions to inspect yet.")
        return

    if state.selected_session not in sessions:
        state.selected_session = sessions[0]

    def on_session_change(e) -> None:
        state.selected_session = e.value
        dive_panel.refresh()

    _render_dive_panel(
        state.sessions,
        state.turns,
        run_id=state.run_id,
        selected_session=state.selected_session,
        on_session_change=on_session_change,
    )


@ui.refreshable
def leaderboard_panel(state: RunState) -> None:
    settings = get_settings()
    rows = load_leaderboard_data(settings.output_dir)
    rows = sorted(rows, key=lambda r: r.get("run_id", ""), reverse=True)
    if not rows:
        ui.label("No runs found yet.")
        return

    def _norm(value: object) -> str:
        if value is None:
            return "unknown"
        text = str(value).strip()
        return text or "unknown"

    agents = sorted({_norm(row.get("Agent")) for row in rows})
    models = sorted({_norm(row.get("Model")) for row in rows})
    benchmarks = sorted({_norm(row.get("Benchmark")) for row in rows})
    subsets = sorted({_norm(row.get("Subset")) for row in rows})

    if not state.selected_agents:
        state.selected_agents = agents
    if not state.selected_models:
        state.selected_models = models
    if not state.selected_benchmarks:
        state.selected_benchmarks = benchmarks
    if not state.selected_subsets:
        state.selected_subsets = subsets

    def on_agents_change(e) -> None:
        state.selected_agents = e.value or []
        leaderboard_panel.refresh()

    def on_models_change(e) -> None:
        state.selected_models = e.value or []
        leaderboard_panel.refresh()

    def on_benchmarks_change(e) -> None:
        state.selected_benchmarks = e.value or []
        leaderboard_panel.refresh()

    def on_subsets_change(e) -> None:
        state.selected_subsets = e.value or []
        leaderboard_panel.refresh()

    def on_min_tasks_change(e) -> None:
        try:
            state.min_tasks = int(e.value)
        except Exception:
            state.min_tasks = 0
        leaderboard_panel.refresh()

    with ui.card().classes("w-full card p-4"):
        with ui.row().classes("w-full"):
            ui.select(
                agents,
                value=state.selected_agents,
                label="Agent",
                multiple=True,
            ).props("dense").on_value_change(on_agents_change)
            ui.select(
                models,
                value=state.selected_models,
                label="Model",
                multiple=True,
            ).props("dense").on_value_change(on_models_change)
            ui.select(
                benchmarks,
                value=state.selected_benchmarks,
                label="Benchmark",
                multiple=True,
            ).props("dense").on_value_change(on_benchmarks_change)
            ui.select(
                subsets,
                value=state.selected_subsets,
                label="Subset",
                multiple=True,
            ).props("dense").on_value_change(on_subsets_change)
            ui.number(
                label="Min Tasks",
                value=state.min_tasks,
                min=0,
                step=1,
            ).props("dense").on_value_change(on_min_tasks_change)

    def _include(row: dict) -> bool:
        if row.get("Agent", "unknown") not in state.selected_agents:
            return False
        if row.get("Model", "unknown") not in state.selected_models:
            return False
        if row.get("Benchmark", "unknown") not in state.selected_benchmarks:
            return False
        if row.get("Subset", "unknown") not in state.selected_subsets:
            return False
        try:
            return int(row.get("Num Tasks", 0)) >= state.min_tasks
        except Exception:
            return False

    filtered = [row for row in rows if _include(row)]
    with ui.card().classes("w-full card p-4"):
        ui.table(columns=LEADERBOARD_COLUMNS, rows=filtered, row_key="run_id").classes("w-full").props("flat")


@ui.refreshable
def history_panel(state: RunState) -> None:
    settings = get_settings()

    def _normalize_history_root(value: str) -> str:
        raw = (value or "").strip()
        if not raw:
            return settings.output_dir
        return os.path.abspath(os.path.expanduser(raw))

    history_root = _normalize_history_root(state.history_root)
    runs = []
    if os.path.isdir(history_root):
        for name in os.listdir(history_root):
            results_path = RunPaths(run_id=name, output_dir=history_root).results
            if os.path.isfile(results_path):
                runs.append(name)
    runs = sorted(runs, reverse=True)

    if state.selected_history_run not in runs:
        state.selected_history_run = runs[0] if runs else None
        state.selected_history_session = None
        state.active_tabs["history"] = "Overview"

    def on_history_change(e) -> None:
        state.selected_history_run = e.value
        state.selected_history_session = None
        state.active_tabs["history"] = "Overview"
        history_panel.refresh()

    def on_history_root_change(e) -> None:
        state.history_root = e.value or ""
        state.selected_history_run = None
        state.selected_history_session = None
        state.active_tabs["history"] = "Overview"
        history_panel.refresh()

    def open_history_browser() -> None:
        browser_state = {"path": history_root}

        dialog = ui.dialog()

        def refresh_entries(container) -> None:
            container.clear()
            current = browser_state["path"]
            with container:
                if not os.path.isdir(current):
                    ui.label("Directory not found.").classes("text-negative")
                    return
                parent = os.path.dirname(current.rstrip(os.sep))
                entries = [name for name in sorted(os.listdir(current)) if os.path.isdir(os.path.join(current, name))]
                if parent and parent != current:
                    entries = ["..", *entries]
                if not entries:
                    ui.label("(empty)")
                    return
                for name in entries:
                    ui.button(
                        name,
                        on_click=lambda n=name: on_entry_click(n, container),
                    ).props("flat dense").classes("justify-start w-full")

        def on_entry_click(name: str, container) -> None:
            current = browser_state["path"]
            if name == "..":
                parent = os.path.dirname(current.rstrip(os.sep))
                if parent and parent != current:
                    browser_state["path"] = parent
            else:
                browser_state["path"] = os.path.join(current, name)
            path_input.value = browser_state["path"]
            refresh_entries(container)

        def on_select() -> None:
            if not os.path.isdir(browser_state["path"]):
                ui.notify("Directory not found.")
                return
            state.history_root = browser_state["path"]
            state.selected_history_run = None
            state.selected_history_session = None
            state.active_tabs["history"] = "Overview"
            history_panel.refresh()
            dialog.close()

        with dialog, ui.card().classes("w-[520px] max-w-full"):
            ui.label("Select run directory")
            with ui.row().classes("w-full items-center gap-2"):
                path_input = (
                    ui.input(
                        label="Directory",
                        value=browser_state["path"],
                    )
                    .props("dense")
                    .style("flex: 1;")
                )

            entries_box = ui.column().classes("w-full gap-2").style("max-height: 320px; overflow-y: auto;")
            refresh_entries(entries_box)

            def on_path_change(e) -> None:
                browser_state["path"] = e.value or ""
                refresh_entries(entries_box)

            path_input.on_value_change(on_path_change)

            with ui.row().classes("w-full justify-end gap-2"):
                ui.button("Cancel", on_click=dialog.close).props("flat dense")
                ui.button("Select", on_click=on_select).props("dense").style(
                    "background:#111111 !important; color:#ffffff !important;"
                )

        dialog.open()

    def on_history_session_change(e) -> None:
        state.selected_history_session = e.value
        state.active_tabs["history"] = "Sessions"
        history_panel.refresh()

    with ui.card().classes("w-full card p-4"):
        with ui.row().classes("w-full items-center gap-3"):
            ui.input(
                label="Run directory",
                value=state.history_root or settings.output_dir,
                placeholder=settings.output_dir,
            ).props("dense").style("min-width: 360px;").on_value_change(on_history_root_change)
            ui.button(icon="folder_open").props("dense flat").on_click(open_history_browser)
        ui.select(runs, value=state.selected_history_run, label="Select run").props("dense").style(
            "min-width: 280px;"
        ).on_value_change(on_history_change)
        if not os.path.isdir(history_root):
            ui.label(f"Directory not found: {history_root}")
        if not state.selected_history_run:
            ui.label("No runs found.")
            return

    run_id = state.selected_history_run
    context = _load_run_context(run_id, output_dir=history_root)
    sessions = _build_history_sessions(run_id, context.results, output_dir=history_root)

    session_ids = sorted(sessions.keys())
    if state.selected_history_session not in session_ids:
        state.selected_history_session = session_ids[0] if session_ids else None

    turns: dict[str, list[dict]] = {}
    if state.selected_history_session:
        turns[state.selected_history_session] = _load_history_turns(
            run_id,
            state.selected_history_session,
            output_dir=history_root,
        )

    with ui.card().classes("w-full card p-4"):
        tabs, tab_by_name, active_tab = _init_tabs(state, "history", ["Overview", "Sessions", "Log"])
        overview_tab = tab_by_name["Overview"]
        sessions_tab = tab_by_name["Sessions"]
        log_tab = tab_by_name["Log"]

        with ui.tab_panels(tabs, value=active_tab).classes("w-full"):
            with ui.tab_panel(overview_tab):

                def _open_history_session(row: dict) -> None:
                    _open_session_from_row(state, "history", row, history_panel.refresh)

                _render_overview_panel(
                    sessions,
                    run_active=False,
                    run_id=run_id,
                    results=context.results,
                    planned_sessions=context.planned_sessions,
                    total_workers=context.total_workers,
                    run_meta=context.run_meta,
                    on_open_session=_open_history_session,
                )
            with ui.tab_panel(sessions_tab):
                _render_dive_panel(
                    sessions,
                    turns,
                    run_id=run_id,
                    selected_session=state.selected_history_session,
                    on_session_change=on_history_session_change,
                    output_dir=history_root,
                )
            with ui.tab_panel(log_tab):
                _render_run_log(run_id, output_dir=history_root)


def _set_enabled(control: Any, enabled: bool) -> None:
    if hasattr(control, "enabled"):
        control.enabled = enabled
        return
    if enabled and hasattr(control, "enable"):
        control.enable()
    elif not enabled and hasattr(control, "disable"):
        control.disable()


def _set_controls_enabled(controls: dict[str, Any], enabled: bool) -> None:
    for data in controls.values():
        control = data[1]
        _set_enabled(control, enabled)


def _set_visible(control: Any, visible: bool) -> None:
    if hasattr(control, "visible"):
        control.visible = visible
        return
    if visible and hasattr(control, "show"):
        control.show()
    elif not visible and hasattr(control, "hide"):
        control.hide()


def build_run_tab(state: RunState) -> RunViews:
    bench_label_to_key, agent_label_to_key = get_display_mappings()
    bench_labels = list(bench_label_to_key.keys())
    agent_labels = list(agent_label_to_key.keys())

    if state.bench_key is None and bench_labels:
        state.bench_key = bench_label_to_key[bench_labels[0]]
    if state.agent_key is None and agent_labels:
        state.agent_key = agent_label_to_key[agent_labels[0]]

    with ui.column().classes("w-full items-center"):
        with ui.column().classes("w-full max-w-6xl gap-4"):
            with ui.element("div").classes("w-full split-grid"):
                with ui.card().classes("w-full card p-4"):
                    with ui.row().classes("w-full items-center justify-between gap-3"):
                        ui.label("Agent")
                        agent_select = (
                            ui.select(
                                agent_labels,
                                value=agent_labels[0] if agent_labels else None,
                                label="",
                            )
                            .props("dense")
                            .style("min-width: 220px;")
                        )
                    with ui.expansion("", value=False).classes("w-full"):
                        agent_form_panel(state)
                with ui.card().classes("w-full card p-4"):
                    with ui.row().classes("w-full items-center justify-between gap-3"):
                        ui.label("Benchmark")
                        bench_select = (
                            ui.select(
                                bench_labels,
                                value=bench_labels[0] if bench_labels else None,
                                label="",
                            )
                            .props("dense")
                            .style("min-width: 220px;")
                        )
                    with ui.expansion("", value=False).classes("w-full"):
                        bench_form_panel(state)

            with ui.card().classes("w-full card p-4"):
                with ui.row().classes("w-full gap-4 items-end"):
                    num_tasks_input = ui.number(
                        label="Num Tasks",
                        value=state.num_tasks or 5,
                        min=0,
                        step=1,
                    ).props("dense")
                    max_workers_input = ui.number(label="Parallel Workers", value=state.max_workers, min=1).props(
                        "dense"
                    )
                    start_button = (
                        ui.button("Start Run")
                        .classes("ml-auto start-run-btn")
                        .style("background:#39ff14 !important; color:#0b0f10 !important;")
                    )

            with ui.card().classes("w-full card p-4") as run_panel_box:
                run_tabs, tab_by_name, active_tab = _init_tabs(state, "run", ["Overview", "Sessions", "Log"])
                overview_tab = tab_by_name["Overview"]
                sessions_tab = tab_by_name["Sessions"]
                log_tab = tab_by_name["Log"]

                with ui.tab_panels(run_tabs, value=active_tab).classes("w-full"):
                    with ui.tab_panel(overview_tab) as overview_panel_el:
                        overview_panel(state)
                    with ui.tab_panel(sessions_tab) as sessions_panel_el:
                        dive_panel(state)
                    with ui.tab_panel(log_tab) as log_panel_el:
                        run_log_panel(state)

    def on_bench_change(e) -> None:
        label = e.value
        state.bench_key = bench_label_to_key.get(label)
        state.bench_controls = {}
        bench_form_panel.refresh()

    def on_agent_change(e) -> None:
        label = e.value
        state.agent_key = agent_label_to_key.get(label)
        agent_form_panel.refresh()

    def on_workers_change(e) -> None:
        try:
            state.max_workers = int(e.value)
        except Exception:
            state.max_workers = 1

    def on_num_tasks_change(e) -> None:
        try:
            value = int(e.value)
        except Exception:
            state.num_tasks = None
            return
        state.num_tasks = value if value > 0 else None

    bench_select.on_value_change(on_bench_change)
    agent_select.on_value_change(on_agent_change)
    num_tasks_input.on_value_change(on_num_tasks_change)
    max_workers_input.on_value_change(on_workers_change)

    show_sessions = state.run_active or bool(state.sessions)
    _set_visible(run_panel_box, show_sessions)
    _set_visible(overview_tab, show_sessions)
    _set_visible(sessions_tab, show_sessions)
    _set_visible(log_tab, show_sessions)
    _set_visible(overview_panel_el, show_sessions)
    _set_visible(sessions_panel_el, show_sessions)
    _set_visible(log_panel_el, show_sessions)

    def start_run() -> None:
        if state.run_active:
            return
        if state.bench_key is None or state.agent_key is None:
            ui.notify("Please select a benchmark and agent.")
            return

        num_tasks_value = None
        try:
            raw = num_tasks_input.value
            if raw is not None:
                parsed = int(raw)
                if parsed > 0:
                    num_tasks_value = parsed
        except Exception:
            num_tasks_value = None
        state.num_tasks = num_tasks_value

        benchmark = None
        agent = None
        try:
            bench_cls = load_benchmark_class(state.bench_key)
            agent_cls = load_agent_class(state.agent_key)
            bench_values = _collect_values(state.bench_controls)
            agent_values = _collect_values(state.agent_controls)
            benchmark = bench_cls(**bench_values)
            agent = agent_cls(**agent_values)
            state.planned_sessions = num_tasks_value if num_tasks_value else None
        except Exception as exc:
            ui.notify(f"Config error: {exc}")
            _LOG.info("config error: %s", exc)
            return
        finally:
            if benchmark is not None:
                try:
                    benchmark.close()
                except Exception:
                    _LOG.info("benchmark preview close failed")
            if agent is not None:
                try:
                    agent.close()
                except Exception:
                    _LOG.info("agent preview close failed")

        settings = get_settings()
        output_dir = settings.output_dir

        state.events = []
        state.sessions = {}
        state.turns = {}
        state.refresh_needed = True

        ctx_mgr = run_scope(output_dir=output_dir)
        ctx = ctx_mgr.__enter__()
        dashboard_observer = DashboardEventsObserver()

        state.tracker = dashboard_observer
        state.run_id = ctx.run_id
        state.run_context_manager = ctx_mgr
        state.run_active = True
        os.environ["EXGENTIC_MAX_WORKERS"] = str(state.max_workers)
        mode = "parallel" if state.max_workers > 1 else "sequential"
        _LOG.info(
            "starting run bench=%s agent=%s mode=%s workers=%s run_id=%s",
            state.bench_key,
            state.agent_key,
            mode,
            state.max_workers,
            state.run_id,
        )

        def worker() -> None:
            try:
                _LOG.info("worker started")
                config = RunConfig(
                    benchmark=state.bench_key,
                    agent=state.agent_key,
                    benchmark_kwargs=bench_values,
                    agent_kwargs=agent_values,
                    output_dir=output_dir,
                    max_workers=state.max_workers if state.max_workers > 1 else None,
                    run_id=state.run_id,
                    num_tasks=num_tasks_value,
                )
                evaluate(
                    config,
                    observers=[dashboard_observer],
                )
                _LOG.info("worker finished")
            except Exception:
                _LOG.exception("worker crashed")

        thread = threading.Thread(target=worker, daemon=True)
        thread.start()
        state.thread = thread

    start_button.on("click", lambda _: start_run())

    return RunViews(
        start_button=start_button,
        bench_select=bench_select,
        agent_select=agent_select,
        num_tasks_input=num_tasks_input,
        max_workers_input=max_workers_input,
        bench_form=bench_form_panel,
        agent_form=agent_form_panel,
        overview_panel=overview_panel,
        sessions_panel=dive_panel,
        run_log_panel=run_log_panel,
        leaderboard_panel=leaderboard_panel,
        history_panel=history_panel,
        run_panel_box=run_panel_box,
        overview_tab=overview_tab,
        sessions_tab=sessions_tab,
        log_tab=log_tab,
        overview_panel_el=overview_panel_el,
        sessions_panel_el=sessions_panel_el,
        log_panel_el=log_panel_el,
    )


def build_leaderboard_tab(state: RunState) -> None:
    leaderboard_panel(state)


def build_history_tab(state: RunState) -> None:
    history_panel(state)


def refresh_ui(state: RunState, views: RunViews) -> None:
    changed = process_new_events(state)

    if state.run_active and state.thread and not state.thread.is_alive():
        state.run_active = False
        if state.run_context_manager is not None:
            try:
                state.run_context_manager.__exit__(None, None, None)
            except Exception:
                pass
            state.run_context_manager = None
        state.refresh_needed = True

    _set_enabled(views.start_button, not state.run_active)
    _set_enabled(views.bench_select, not state.run_active)
    _set_enabled(views.agent_select, not state.run_active)
    _set_enabled(views.num_tasks_input, not state.run_active)
    _set_enabled(views.max_workers_input, not state.run_active)
    _set_controls_enabled(state.bench_controls, not state.run_active)
    _set_controls_enabled(state.agent_controls, not state.run_active)

    show_sessions = state.run_active or bool(state.sessions)
    _set_visible(views.run_panel_box, show_sessions)
    _set_visible(views.overview_tab, show_sessions)
    _set_visible(views.sessions_tab, show_sessions)
    _set_visible(views.log_tab, show_sessions)
    _set_visible(views.overview_panel_el, show_sessions)
    _set_visible(views.sessions_panel_el, show_sessions)
    _set_visible(views.log_panel_el, show_sessions)

    if state.last_run_active != state.run_active:
        views.bench_form.refresh()
        views.agent_form.refresh()
        state.last_run_active = state.run_active

    if changed:
        views.overview_panel.refresh()
        views.sessions_panel.refresh()
        views.run_log_panel.refresh()

    if state.refresh_needed:
        views.leaderboard_panel.refresh()
        views.history_panel.refresh()
        views.run_log_panel.refresh()
        state.refresh_needed = False



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from .controller import Controller, CoreController, LimitController
from .observer import Observer
from .run import core_aggregate, core_evaluate, core_execute
from .session import run_session
from .termination import (
    AgentError,
    AgentTerminationError,
    BenchmarkError,
    BenchmarkTerminationError,
    InvalidActionError,
    InvalidObservationError,
    RunCancelError,
    SessionCancelError,
    SessionLimitReachedError,
)
from .tracker import Tracker

__all__ = [
    "AgentError",
    "AgentTerminationError",
    "BenchmarkError",
    "BenchmarkTerminationError",
    "Controller",
    "CoreController",
    "LimitController",
    "InvalidActionError",
    "InvalidObservationError",
    "Observer",
    "RunCancelError",
    "Tracker",
    "SessionCancelError",
    "SessionLimitReachedError",
    "SessionTermination",
    "run_session",
    "core_aggregate",
    "core_execute",
    "core_evaluate",
]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from exgentic.agents.litellm_tool_calling.utils import ToolsActionsRegistry
from exgentic.core.actions import ActionsHandler
from exgentic.core.types import SingleAction
from pydantic import BaseModel


class EmptyArgs(BaseModel):
    pass


class DummyAction(SingleAction):
    arguments: EmptyArgs


def test_unknown_tool_call_name_yields_unknown_action_observation():
    registry = ToolsActionsRegistry(actions=[])
    tool_calls = [{"name": "not_a_tool", "arguments": "{}", "id": "call-1"}]

    action = registry.tool_calls_to_action(tool_calls)

    assert action is not None
    assert action.validation.name_valid is False
    assert action.validation.error == "Unknown action"

    handler = ActionsHandler()
    observation = handler.execute(action)

    assert "Unknown action" in str(observation.result)



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import asyncio
import logging
from typing import Any

import httpx
from agents import Agent as OpenAIAgent
from agents import Runner
from agents.extensions.models.litellm_model import LitellmModel
from agents.lifecycle import RunHooksBase
from agents.mcp import MCPServerStreamableHttp, MCPServerStreamableHttpParams
from agents.model_settings import ModelSettings as OpenAIModelSettings
from agents.model_settings import Reasoning
from agents.run import RunConfig
from agents.usage import Usage

from ...adapters.agents.mcp_agent import MCPAgentInstance
from ...core.context import get_context
from ...core.types import ModelSettings, RetryStrategy
from ...integrations.litellm.health import acheck_model_accessible
from ...observers.logging import (
    attach_library_logger_to_handler,
    restore_library_logger,
)
from ...utils.cost import UpdatableCostReport, litellm_tokens_cost
from ...utils.settings import get_settings
from ...utils.sync import run_sync
from .openai_mcp_agent import MCPConfig

settings = get_settings()


class _UsageRunHooks(RunHooksBase[dict[str, Any], OpenAIAgent]):
    def __init__(self, record_usage) -> None:
        super().__init__()
        self._record_usage = record_usage

    async def on_llm_end(self, context, agent, response) -> None:
        self._record_usage(response.usage)


class RetryingLitellmModel(LitellmModel):
    def __init__(
        self,
        model: str,
        *,
        num_retries: int,
        retry_after: float,
        retry_strategy: str | RetryStrategy,
    ):
        super().__init__(model=model)
        if isinstance(retry_strategy, RetryStrategy):
            retry_strategy = retry_strategy.value
        if retry_strategy not in ("exponential_backoff_retry", "constant_retry"):
            raise ValueError(f"Unsupported retry_strategy: {retry_strategy}")
        if num_retries < 0:
            raise ValueError("num_retries must be >= 0")
        if retry_after < 0:
            raise ValueError("retry_after must be >= 0")
        self._num_retries = num_retries
        self._retry_after = retry_after
        self._retry_strategy = retry_strategy

    async def _fetch_response(self, *args, **kwargs):
        # Inject context for OTEL tracing via model_settings.metadata
        # The parent class extracts metadata from model_settings and passes it to litellm.acompletion()
        # Note: metadata must be Dict[str, str], so we serialize Context fields individually
        ctx = get_context()

        # model_settings is the 3rd positional argument (index 2)
        if len(args) > 2:
            model_settings = args[2]
            if model_settings.metadata is None:
                model_settings.metadata = {}

            # Serialize Context fields as individual string metadata entries
            model_settings.metadata["exgentic_ctx_run_id"] = ctx.run_id
            model_settings.metadata["exgentic_ctx_output_dir"] = ctx.output_dir
            model_settings.metadata["exgentic_ctx_cache_dir"] = ctx.cache_dir
            if ctx.session_id is not None:
                model_settings.metadata["exgentic_ctx_session_id"] = ctx.session_id
            if ctx.task_id is not None:
                model_settings.metadata["exgentic_ctx_task_id"] = ctx.task_id
            model_settings.metadata["exgentic_ctx_role"] = ctx.role.value
            if ctx.otel_context is not None:
                model_settings.metadata["exgentic_ctx_otel_trace_id"] = ctx.otel_context.trace_id
                model_settings.metadata["exgentic_ctx_otel_span_id"] = ctx.otel_context.span_id

        for attempt in range(self._num_retries + 1):
            try:
                return await super()._fetch_response(*args, **kwargs)
            except Exception as exc:
                if attempt >= self._num_retries:
                    raise
                delay = self._retry_after
                if self._retry_strategy == RetryStrategy.EXPONENTIAL_BACKOFF.value:
                    delay *= 2**attempt
                logging.getLogger(__name__).warning(
                    "OpenAI MCP LiteLLM call failed (attempt %d/%d): %s",
                    attempt + 1,
                    self._num_retries + 1,
                    exc,
                )
                if delay > 0:
                    await asyncio.sleep(delay)
        return None


class OpenAIMCPAgentInstance(MCPAgentInstance):
    """OpenAI Agents SDK + MCP (sync entrypoint, async core)."""

    def __init__(
        self,
        session_id: str,
        model_id: str,
        max_steps: int = 150,
        model_settings: ModelSettings | None = None,
        mcp_config: MCPConfig | dict | None = None,
    ):
        super().__init__(session_id)
        self.model_id = model_id
        self.max_steps = max_steps
        if model_settings is None:
            self.model_settings = ModelSettings()
        elif isinstance(model_settings, ModelSettings):
            self.model_settings = model_settings
        else:
            raise ValueError("model_settings must be a ModelSettings instance.")
        if mcp_config is None:
            self.mcp_config = MCPConfig()
        elif isinstance(mcp_config, dict):
            self.mcp_config = MCPConfig(**mcp_config)
        else:
            self.mcp_config = mcp_config
        self._total_input_tokens = 0
        self._total_output_tokens = 0
        self._model_access_checked = False

    async def _check_model_access_once(self) -> None:
        if self._model_access_checked or self.mcp_config.skip_health_check:
            return
        self.logger.info("Running LiteLLM model health check (model=%s)", self.model_id)
        await acheck_model_accessible(self.model_id)
        self._model_access_checked = True

    def _record_usage(self, usage: Usage | None) -> None:
        if usage is None:
            return
        self._total_input_tokens += usage.input_tokens
        self._total_output_tokens += usage.output_tokens

    def run_mcp_agent(self, mcp_host: str, mcp_port: int) -> Any:
        # Run async core on the shared loop (sync API)
        return run_sync(self.run_mcp_agent_async(mcp_host, mcp_port), timeout=600.0)

    async def run_mcp_agent_async(self, mcp_host: str, mcp_port: int) -> Any:
        RunConfig.tracing_disabled = True

        prompt = self._build_prompt()

        file_handler = next(
            (h for h in self.logger.handlers if isinstance(h, logging.FileHandler)),
            None,
        )
        logger_states: list[tuple] = []

        try:
            if file_handler:
                logger_states += [
                    attach_library_logger_to_handler("agents", file_handler),
                    attach_library_logger_to_handler(__name__, file_handler),
                ]
                logging.getLogger("agents").setLevel(logging.INFO)
            await self._check_model_access_once()

            # Create custom httpx client factory with extended timeout
            def httpx_client_factory(headers=None, timeout=None, auth=None):
                if (
                    self.mcp_config.http_timeout_seconds is None
                    and self.mcp_config.sse_read_timeout_seconds is None
                    and self.mcp_config.http_connect_timeout_seconds is None
                ):
                    client_timeout = httpx.Timeout(None)
                else:
                    client_timeout = httpx.Timeout(
                        self.mcp_config.http_timeout_seconds,
                        connect=self.mcp_config.http_connect_timeout_seconds,
                        read=self.mcp_config.sse_read_timeout_seconds,
                    )
                return httpx.AsyncClient(
                    headers=headers,
                    timeout=client_timeout,
                    auth=auth,
                )

            mcp_params: dict[str, Any] = {
                "url": f"http://{mcp_host}:{mcp_port}/mcp",
                "httpx_client_factory": httpx_client_factory,
                "terminate_on_close": self.mcp_config.terminate_on_close,
            }
            if self.mcp_config.headers is not None:
                mcp_params["headers"] = self.mcp_config.headers
            if self.mcp_config.http_timeout_seconds is not None:
                mcp_params["timeout"] = self.mcp_config.http_timeout_seconds
            if self.mcp_config.sse_read_timeout_seconds is not None:
                mcp_params["sse_read_timeout"] = self.mcp_config.sse_read_timeout_seconds

            async with MCPServerStreamableHttp(
                params=MCPServerStreamableHttpParams(**mcp_params),
                cache_tools_list=self.mcp_config.cache_tools_list,
                name=self.mcp_config.name,
                client_session_timeout_seconds=self.mcp_config.client_session_timeout_seconds,
                use_structured_content=self.mcp_config.use_structured_content,
                max_retry_attempts=self.mcp_config.max_retry_attempts,
                retry_backoff_seconds_base=self.mcp_config.retry_backoff_seconds_base,
                message_handler=self.mcp_config.message_handler,
            ) as mcp_server:
                temperature = self.model_settings.temperature
                reasoning_effort = self.model_settings.reasoning_effort
                openai_model_settings = OpenAIModelSettings(
                    temperature=temperature if temperature is not None else 1.0,
                    max_tokens=self.model_settings.max_tokens,
                    top_p=self.model_settings.top_p,
                    reasoning=(Reasoning(effort=reasoning_effort) if reasoning_effort is not None else None),
                )
                num_retries = self.model_settings.num_retries or 0
                retry_after = self.model_settings.retry_after
                retry_strategy = self.model_settings.retry_strategy.value
                openai_model_settings.extra_args = {
                    "caching": settings.litellm_caching,
                    "max_retries": 0 if num_retries > 0 else 5,
                }
                agent = OpenAIAgent(
                    name="Assistant",
                    instructions=prompt,
                    model=RetryingLitellmModel(
                        model=self.model_id,
                        num_retries=num_retries,
                        retry_after=retry_after,
                        retry_strategy=retry_strategy,
                    ),
                    model_settings=openai_model_settings,
                    mcp_servers=[mcp_server],
                )
                self.logger.info(
                    "Starting OpenAI MCP agent run (model=%s, task=%s, max_turns=%s)",
                    self.model_id,
                    self.task,
                    self.max_steps,
                )
                hooks = _UsageRunHooks(self._record_usage)
                try:
                    result = await Runner.run(
                        agent,
                        self.task,
                        max_turns=self.max_steps,
                        run_config=RunConfig(tracing_disabled=True),
                        hooks=hooks,
                    )
                except Exception:
                    self.logger.exception("OpenAI MCP agent run failed")
                    raise
                if self._total_input_tokens == 0 and self._total_output_tokens == 0:
                    for resp in result.raw_responses:
                        self._record_usage(resp.usage)
                self.logger.info("OpenAI MCP agent run finished: %s", result)
                return result
        finally:
            for state in logger_states:
                if state:
                    restore_library_logger(*state)
            if file_handler:
                file_handler.flush()

    def get_cost(self) -> UpdatableCostReport:
        report = UpdatableCostReport.initialize_empty(model_name=self.model_id)
        if self._total_input_tokens == 0 and self._total_output_tokens == 0:
            return report

        cost = litellm_tokens_cost(
            model_name=self.model_id,
            input_tokens=self._total_input_tokens,
            output_tokens=self._total_output_tokens,
        ).total_cost
        report.add_cost(cost)
        return report

    def _build_prompt(self) -> str:
        prompt = ""
        if self.context:
            prompt += f"Context: {self.context}\n\n"

        prompt += (
            "Complete this task using the available tools. Each tool corresponds to an action "
            "you can take in the environment. Do not respond or ask clarification questions "
            "unless done through a dedicated tool, and only if such tool exist. "
            "Any plain message that is not a tool call will end the run in failure.\n"
        )

        if self.initial_observation is not None and not self.initial_observation.is_empty():
            text = str(self.initial_observation).strip()
            if text:
                prompt += f"\nFirst Observation: {text}\n"

        return prompt



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

import rich_click as click

from ...lib.api import list_agents, list_benchmarks, list_subsets, list_tasks
from ..options import apply_debug_mode
from ..render import render_list, render_named_list


@click.group("list")
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug mode (sets settings.debug=true and log level to DEBUG)",
)
def list_cmd(debug: bool) -> None:
    """List available resources."""
    apply_debug_mode(debug)


@list_cmd.command("benchmarks")
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug mode (sets settings.debug=true and log level to DEBUG)",
)
@click.option("--format", "output_format", type=click.Choice(["text", "json"]), default="text")
def list_benchmarks_cmd(debug: bool, output_format: str) -> None:
    """List available benchmarks."""
    apply_debug_mode(debug)
    items = list_benchmarks()
    render_named_list(items, output_format, title="Benchmarks")


@list_cmd.command("agents")
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug mode (sets settings.debug=true and log level to DEBUG)",
)
@click.option("--format", "output_format", type=click.Choice(["text", "json"]), default="text")
def list_agents_cmd(debug: bool, output_format: str) -> None:
    """List available agents."""
    apply_debug_mode(debug)
    items = list_agents()
    render_named_list(items, output_format, title="Agents")


@list_cmd.command("subsets")
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug mode (sets settings.debug=true and log level to DEBUG)",
)
@click.option("--benchmark", required=True, help="Benchmark slug_name")
@click.option("--format", "output_format", type=click.Choice(["text", "json"]), default="text")
def list_subsets_cmd(debug: bool, benchmark: str, output_format: str) -> None:
    """List available subsets for a benchmark."""
    apply_debug_mode(debug)
    try:
        subsets = list_subsets(benchmark)
    except Exception as exc:
        raise click.ClickException(str(exc)) from exc
    render_list(subsets, output_format, title=f"Subsets ({benchmark})")


@list_cmd.command("tasks")
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug mode (sets settings.debug=true and log level to DEBUG)",
)
@click.option("--benchmark", required=True, help="Benchmark slug_name")
@click.option("--subset", help="Benchmark subset name")
@click.option("--limit", type=int, help="Limit output to first N tasks")
@click.option("--format", "output_format", type=click.Choice(["text", "json"]), default="text")
def list_tasks_cmd(
    debug: bool,
    benchmark: str,
    subset: str | None,
    limit: int | None,
    output_format: str,
) -> None:
    """List task ids for a benchmark."""
    apply_debug_mode(debug)
    try:
        tasks = list_tasks(benchmark=benchmark, subset=subset)
    except Exception as exc:
        raise click.ClickException(str(exc)) from exc
    if limit is not None:
        tasks = tasks[: int(limit)]
    label = f"Tasks ({benchmark})" if subset is None else f"Tasks ({benchmark}:{subset})"
    render_list(tasks, output_format, title=label)


__all__ = [
    "list_cmd",
    "list_benchmarks_cmd",
    "list_agents_cmd",
    "list_subsets_cmd",
    "list_tasks_cmd",
]



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import json
import re
from pathlib import Path
from typing import Any

from pydantic import Field

from ...core.types import SessionScore as BaseSessionScore
from ...utils.paths import SessionPaths


class SessionScore(BaseSessionScore):
    instance_id: str = ""
    agent: dict[str, Any] = Field(default_factory=dict)
    patch: dict[str, Any] = Field(default_factory=dict)
    container: dict[str, Any] = Field(default_factory=dict)
    evaluation: dict[str, Any] = Field(default_factory=dict)
    summary: dict[str, Any] = Field(default_factory=dict)


FILE_OPS_RE = re.compile(r"\b(cp|mv|rm|mkdir|touch|tee|patch)\b")
BUILD_PHASES = {"base": "build_base", "env": "build_env", "instances": "build_instance"}


def build_score(
    paths: SessionPaths,
    num_actions: int,
    max_interactions: int,
    instance_id: str,
    harness_data: Any | None = None,
) -> SessionScore:
    score = SessionScore(score=0.0, success=False, instance_id=instance_id)
    score.agent = _parse_agent(paths)
    score.patch = _parse_patch(paths)
    score.container, score.evaluation = _parse_harness_files(paths, score.patch)
    if harness_data:
        _apply_harness_data(score, harness_data, instance_id)
    score.summary = _build_summary(score, num_actions, max_interactions)
    score.score = score.summary["score"]
    return score


def _parse_agent(paths: SessionPaths) -> dict[str, Any]:
    log_path = paths.benchmark_dir / "session.log"
    if not log_path.exists():
        return {"commands": None, "edit_commands": None, "call_submit": False}
    content = log_path.read_text()
    commands = re.findall(r"\| command: (.+?)(?=\n(?:INFO|ERROR|\[LiteLM\])|$)", content, re.DOTALL)
    edit_cmds = [c for c in commands if _is_edit_cmd(c)]
    return {
        "commands": commands or None,
        "edit_commands": edit_cmds or None,
        "call_submit": bool(re.search(r"submit_patch\s*\| summary:", content)),
    }


def _parse_patch(paths: SessionPaths) -> dict[str, Any]:
    pred_path = paths.benchmark_dir / "predictions.jsonl"
    _none = {"generated": None, "length": None, "structurally_valid": None}
    if not pred_path.exists():
        return _none
    try:
        lines = pred_path.read_text().splitlines()
        if not lines:
            return _none
        pred = json.loads(lines[0])
        patch_len = len(pred.get("model_patch", ""))
        return {
            "generated": patch_len > 0,
            "length": patch_len,
            "structurally_valid": pred.get("patch_valid", False),
        }
    except (json.JSONDecodeError, IndexError):
        return _none


def _parse_harness_files(paths: SessionPaths, patch: dict) -> tuple:
    container = {
        "required": None,
        "build_base": None,
        "build_env": None,
        "build_instance": None,
        "started": None,
        "patch_exists": None,
        "applying_patch": None,
        "patch_applied": None,
        "removed": None,
    }
    evaluation = {"grading": None, "resolved": None, "test_results": {}}

    container["required"] = bool(patch.get("generated") and patch.get("structurally_valid"))
    if not container["required"]:
        return container, evaluation

    for path in paths.benchmark_dir.rglob("*"):
        if path.name == "build_image.log" and "build_images" in str(path):
            try:
                phase = path.parts[-3]
                if phase in BUILD_PHASES:
                    container[BUILD_PHASES[phase]] = path.read_text().strip().endswith("Image built successfully!")
            except IndexError:
                pass

        elif path.name == "run_instance.log" and "run_evaluation" in str(path):
            content = path.read_text()
            if re.search(r"Container .* started", content):
                container["started"] = True
            if re.search(r"Intermediate patch for .* written to logs", content):
                container["patch_exists"] = True
            if re.search(r"now applying to container", content):
                container["applying_patch"] = True
            if re.search(r"Patch Apply Failed", content):
                container["patch_applied"] = False
            if re.search(r"Grading answer for .*", content):
                evaluation["grading"] = True
            if re.search(r"Container .* removed\.", content):
                container["removed"] = True
            if container.get("applying_patch") and container.get("patch_applied") is None:
                container["patch_applied"] = True

        elif path.name == "report.json":
            try:
                report = json.loads(path.read_text())
                if len(report) == 1:
                    data = next(iter(report.values()))
                    container["patch_exists"] = data["patch_exists"]
                    container["patch_applied"] = data["patch_successfully_applied"]
                    evaluation["resolved"] = data["resolved"]
                    for case, cd in data.get("tests_status", {}).items():
                        s, f = len(cd.get("success", [])), len(cd.get("failure", []))
                        if s + f > 0:
                            rate = round(100 * s / (s + f), 2)
                            evaluation["test_results"][case] = {
                                "expected": s + f,
                                "success": s,
                                "failure": f,
                                "rate": rate,
                                "display": f"expected {s + f}: success {s}, failure {f}. success rate: {rate}%",
                            }
            except json.JSONDecodeError:
                pass

    return container, evaluation


def _apply_harness_data(score: SessionScore, hd: Any, instance_id: str):
    if hd.harness_report:
        submitted = set(hd.harness_report.get("submitted_ids", []))
        if submitted != {instance_id}:
            raise ValueError(f"Instance ID mismatch in harness report: expected {{'{instance_id}'}}, got {submitted}")
        if hd.harness_report.get("resolved_instances", 0) > 0:
            score.evaluation["resolved"] = True
    score.success = hd.error is None


def _build_summary(score: SessionScore, num_actions: int, max_interactions: int) -> dict[str, Any]:
    c, e, a, p = score.container, score.evaluation, score.agent, score.patch
    f2p = e.get("test_results", {}).get("FAIL_TO_PASS", {})
    p2p = e.get("test_results", {}).get("PASS_TO_PASS", {})
    final = int(e.get("resolved") or False)
    return {
        "num_actions": num_actions,
        "actions_limit": max_interactions,
        "num_edit_commands": len(a.get("edit_commands") or []),
        "agent_call_submit": 1 if a.get("call_submit") else 0,
        "patch_non_empty": 1 if p.get("generated") else 0,
        "container_status": -1 if not c.get("required") else (1 if c.get("started") else 0),
        "fail_to_pass_rate": f2p.get("rate") if isinstance(f2p, dict) else None,
        "pass_to_pass_rate": p2p.get("rate") if isinstance(p2p, dict) else None,
        "score": final,
    }


def _is_edit_cmd(cmd: str) -> bool:
    c = cmd.lower()
    return (
        ("python" in c and any(p in c for p in ["write_text", "write_bytes", ".write(", "replace(", "open("]))
        or "git apply" in c
        or "sed -i" in c
        or bool(FILE_OPS_RE.search(c))
        or ">>" in c
    )


def write_results(results_path: Path, score: SessionScore):
    results_path.parent.mkdir(parents=True, exist_ok=True)
    results_path.write_text(score.model_dump_json(indent=2))



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

import os
from pathlib import Path


def _resolve_tau2_data_dir() -> str:
    """Return the tau2 data directory path.

    Checks the cache directory first (populated by ``setup.sh``), then falls
    back to the legacy ``installation/`` path for backwards compatibility.
    """
    from ...utils.cache import benchmark_cache_dir

    cache_data = benchmark_cache_dir("tau2") / "data"
    if cache_data.is_dir():
        return str(cache_data)

    # Legacy path (pre-setup.sh installs that cloned into the package tree)
    legacy = Path(__file__).resolve().parent / "installation" / "tau2-bench" / "data"
    return str(legacy)


if "TAU2_DATA_DIR" not in os.environ:
    os.environ["TAU2_DATA_DIR"] = _resolve_tau2_data_dir()



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Tests for LiteLLM health check error handling."""

from __future__ import annotations

import logging
from unittest.mock import patch

import pytest
from exgentic.integrations.litellm.health import check_model_accessible_sync


class MockLiteLLMError(Exception):
    """Mock exception that mimics LiteLLM exceptions with .message attribute."""

    def __init__(self, message: str):
        self.message = message
        super().__init__()

    def __str__(self) -> str:
        """Return empty string to simulate LiteLLM exceptions that don't implement __str__."""
        return ""


def test_health_check_extracts_message_attribute_from_exception(caplog):
    """Test that health check extracts error details from exception.message attribute."""
    caplog.set_level(logging.ERROR)

    with patch("exgentic.utils.sync.run_sync") as mock_run_sync:
        exc = MockLiteLLMError("API key authentication failed")
        mock_run_sync.side_effect = exc

        logger = logging.getLogger("test")

        with pytest.raises(RuntimeError) as exc_info:
            check_model_accessible_sync("test-model", logger)

        error_msg = str(exc_info.value)
        assert "API key authentication failed" in error_msg
        assert "test-model" in error_msg
        assert error_msg == "Model test-model is not accessible: API key authentication failed"


def test_health_check_falls_back_to_str_when_no_message_attribute(caplog):
    """Test that health check falls back to str(exc) when .message is not available."""
    caplog.set_level(logging.ERROR)

    with patch("exgentic.utils.sync.run_sync") as mock_run_sync:
        exc = ValueError("Standard error message")
        mock_run_sync.side_effect = exc

        logger = logging.getLogger("test")

        with pytest.raises(RuntimeError) as exc_info:
            check_model_accessible_sync("test-model", logger)

        error_msg = str(exc_info.value)
        assert "Standard error message" in error_msg
        assert "test-model" in error_msg


def test_health_check_uses_repr_as_last_resort(caplog):
    """Test that health check uses repr(exc) when both .message and str(exc) are empty."""
    caplog.set_level(logging.ERROR)

    class EmptyError(Exception):
        """Exception that returns empty string from __str__."""

        def __str__(self) -> str:
            return ""

    with patch("exgentic.utils.sync.run_sync") as mock_run_sync:
        exc = EmptyError("hidden")
        mock_run_sync.side_effect = exc

        logger = logging.getLogger("test")

        with pytest.raises(RuntimeError) as exc_info:
            check_model_accessible_sync("test-model", logger)

        error_msg = str(exc_info.value)
        assert "EmptyError" in error_msg
        assert "test-model" in error_msg



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from smolagents import ToolCallingAgent
from smolagents.tools import Tool
from smolagents.utils import AgentError

from .base_instance import SmolagentBaseAgentInstance


class SmolagentToolCallingAgentInstance(SmolagentBaseAgentInstance):
    """Smolagent implementation."""

    def run_smolagent(self, tools: list[Tool]):
        self._agent = ToolCallingAgent(
            tools=tools,
            model=self.get_internal_model(),
            # use_structured_outputs_internally=True,
            logger=self.get_smolagent_logger(),
        )

        prompt = f"Task: {self.task}\n\n"
        if self.context:
            prompt += f"Context: {self.context}\n\n"
        prompt += (
            "Complete this task using the available tools. "
            "Each tool corresponds to an action you can take in the environment.\n"
        )
        if self.initial_observation is not None and not self.initial_observation.is_empty():
            text = str(self.initial_observation).strip()
            if text:
                prompt += f"\nFirst Observation: {text}\n"
        try:
            self._agent.run(task=prompt, max_steps=self.max_steps)
        except AgentError as e:
            self.logger.info(f"AgentError: {e}")
            raise



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

"""Tests for :mod:`exgentic.adapters.runners._utils`."""

from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

from exgentic.adapters.runners._utils import find_project_root


def test_find_project_root_returns_repo_root():
    """When a pyproject.toml exists in a parent, return that directory."""
    root = find_project_root()
    assert (root / "pyproject.toml").exists()


def test_find_project_root_falls_back_to_dot_exgentic(tmp_path: Path):
    """When no pyproject.toml is found, fall back to ~/.exgentic/."""
    fake_home = tmp_path / "home"
    fake_home.mkdir()

    # Create a fake __file__ path with no pyproject.toml in any parent
    fake_file = tmp_path / "lib" / "pkg" / "mod.py"
    fake_file.parent.mkdir(parents=True)
    fake_file.touch()

    with (
        patch(
            "exgentic.adapters.runners._utils.Path.__file__",
            create=True,
        ),
        patch(
            "exgentic.adapters.runners._utils.Path.home",
            return_value=fake_home,
        ),
    ):
        # Patch __file__ at the module level so Path(__file__) resolves
        # to a location without pyproject.toml in any ancestor.
        import exgentic.adapters.runners._utils as mod

        original_file = mod.__file__
        try:
            mod.__file__ = str(fake_file)
            result = find_project_root()
        finally:
            mod.__file__ = original_file

    expected = fake_home / ".exgentic"
    assert result == expected
    assert expected.is_dir()


def test_find_project_root_fallback_is_idempotent(tmp_path: Path):
    """Calling find_project_root twice with fallback doesn't error."""
    fake_home = tmp_path / "home"
    fake_home.mkdir()

    fake_file = tmp_path / "lib" / "mod.py"
    fake_file.parent.mkdir(parents=True)
    fake_file.touch()

    with patch(
        "exgentic.adapters.runners._utils.Path.home",
        return_value=fake_home,
    ):
        import exgentic.adapters.runners._utils as mod

        original_file = mod.__file__
        try:
            mod.__file__ = str(fake_file)
            result1 = find_project_root()
            result2 = find_project_root()
        finally:
            mod.__file__ = original_file

    assert result1 == result2
    assert result1 == fake_home / ".exgentic"



# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2026, The Exgentic organization and its contributors.

from __future__ import annotations

from pathlib import Path
from typing import Optional

from ..base import BaseCLIConfig, BaseCLIWrapper, ExecutionBackend


class CodexCLIConfig(BaseCLIConfig):
    model_id: str
    env_key: str = "OPENAI_API_KEY"
    profile: str = "temp_model"
    provider_name: str = "Provider"
    image: str = "exgentic-codex:dev"


class CodexCLI(BaseCLIWrapper):
    """Standalone wrapper for launching Codex CLI."""

    def __init__(
        self,
        env: Optional[dict[str, str]] = None,
        log_path: Optional[Path] = None,
        logger=None,
        runner: ExecutionBackend = ExecutionBackend.PROCESS,
    ) -> None:
        super().__init__(env=env, log_path=log_path, config_dir=None, logger=logger, runner=runner)
        self.config_prefix = "codex_cli_"
        self.spawn_error_message = "Failed to start Codex CLI"

    def build_env(self, *, cfg_root: Path, prompt: str, config: CodexCLIConfig) -> dict[str, str]:
        env = self.env.copy()
        env["OPENAI_API_BASE"] = config.provider_url
        api_key = env.get(config.env_key) or "dummy-api-key"
        env[config.env_key] = api_key
        return env

    def build_command(self, *, cfg_root: Path, prompt: str, config: CodexCLIConfig) -> list[str]:
        # Rewrite localhost addresses to host gateway for container runners
        mcp_host = config.mcp_host
        provider_url = config.provider_url
        from ..command_runner import ContainerRunner

        if isinstance(self.runner, ContainerRunner):
            if mcp_host in ("0.0.0.0", "127.0.0.1", "localhost"):
                mcp_host = self.runner.host_gateway
            for local in ("://127.0.0.1:", "://localhost:"):
                if local in provider_url:
                    provider_url = provider_url.replace(local, f"://{self.runner.host_gateway}:")
                    break

        mcp_url = f"http://{mcp_host}:{config.mcp_port}/mcp"
        overrides = [
            f'mcp_servers.environment.url="{mcp_url}"',
            f'model_providers.temp.name="{config.provider_name}"',
            f'model_providers.temp.base_url="{provider_url}"',
            f'model_providers.temp.env_key="{config.env_key}"',
            f'profiles.{config.profile}.model_provider="temp"',
            f'profiles.{config.profile}.model="{config.model_id}"',
        ]

        cmd: list[str] = ["codex", "exec", "--skip-git-repo-check", "--profile", config.profile]
        for override in overrides:
            cmd.extend(["-c", override])
        cmd.append(prompt)

        return cmd
